import inspect
import asyncio
from dataclasses import dataclass
from typing import (
    Callable,
    Optional,
    List,
    Dict,
    Any,
    Union,
    Type,
    get_origin,
    get_args,
    TYPE_CHECKING,
    Literal,
    Annotated,
    TypeVar,
    cast,
)

from .commands import (
    SlashCommand,
    UserCommand,
    MessageCommand,
    AppCommand,
    AppCommandGroup,
)
from .hybrid import HybridCommand
from disagreement.interactions import (
    ApplicationCommandOption,
    ApplicationCommandOptionChoice,
    Snowflake,
)
from disagreement.enums import (
    ApplicationCommandOptionType,
    IntegrationType,
    InteractionContextType,
    # Assuming ChannelType will be added to disagreement.enums
)

if TYPE_CHECKING:
    from disagreement.client import Client  # For potential future use
    from disagreement.models import Channel, User

    # Assuming TextChannel, VoiceChannel etc. might be defined or aliased
    # For now, we'll use string comparisons for channel types or rely on a yet-to-be-defined ChannelType enum
    Channel = Any
    Member = Any
    Role = Any
    Attachment = Any
    # from .cog import Cog # Placeholder
else:
    # Runtime fallbacks for optional model classes
    from disagreement.models import Channel

    Client = Any  # type: ignore
    User = Any  # type: ignore
    Member = Any  # type: ignore
    Role = Any  # type: ignore
    Attachment = Any  # type: ignore

# Mapping Python types to Discord ApplicationCommandOptionType
# This will need to be expanded and made more robust.
# Consider using a registry or a more sophisticated type mapping system.
_type_mapping: Dict[Any, ApplicationCommandOptionType] = (
    {  # Changed Type to Any for key due to placeholders
        str: ApplicationCommandOptionType.STRING,
        int: ApplicationCommandOptionType.INTEGER,
        bool: ApplicationCommandOptionType.BOOLEAN,
        float: ApplicationCommandOptionType.NUMBER,  # Discord 'NUMBER' type is for float/double
        User: ApplicationCommandOptionType.USER,
        Channel: ApplicationCommandOptionType.CHANNEL,
        # Placeholders for actual model types from disagreement.models
        # These will be resolved to their actual types when TYPE_CHECKING is False or via isinstance checks
    }
)


# Helper dataclass for storing extra option metadata
@dataclass
class OptionMetadata:
    channel_types: Optional[List[int]] = None
    min_value: Optional[Union[int, float]] = None
    max_value: Optional[Union[int, float]] = None
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    autocomplete: bool = False


# Ensure these are updated if model names/locations change
if TYPE_CHECKING:
    # _type_mapping[User] = ApplicationCommandOptionType.USER # Already added above
    _type_mapping[Member] = ApplicationCommandOptionType.USER  # Member implies User
    _type_mapping[Role] = ApplicationCommandOptionType.ROLE
    _type_mapping[Attachment] = ApplicationCommandOptionType.ATTACHMENT
    _type_mapping[Channel] = ApplicationCommandOptionType.CHANNEL

# TypeVar for the app command decorator factory
AppCmdType = TypeVar("AppCmdType", bound=AppCommand)


def _extract_options_from_signature(
    func: Callable[..., Any], option_meta: Optional[Dict[str, OptionMetadata]] = None
) -> List[ApplicationCommandOption]:
    """
    Inspects a function signature and generates ApplicationCommandOption list.
    """
    options: List[ApplicationCommandOption] = []
    params = inspect.signature(func).parameters

    doc = inspect.getdoc(func)
    param_descriptions: Dict[str, str] = {}
    if doc:
        for line in inspect.cleandoc(doc).splitlines():
            line = line.strip()
            if line.startswith(":param"):
                try:
                    _, rest = line.split(" ", 1)
                    name, desc = rest.split(":", 1)
                    param_descriptions[name.strip()] = desc.strip()
                except ValueError:
                    continue

    # Skip 'self' (for cogs) and 'ctx' (context) parameters
    param_iter = iter(params.values())
    first_param = next(param_iter, None)

    # Heuristic: if the function is bound to a class (cog), 'self' might be the first param.
    # A more robust way would be to check if `func` is a method of a Cog instance later.
    # For now, simple name check.
    if first_param and first_param.name == "self":
        first_param = next(param_iter, None)  # Consume 'self', get next

    if first_param and first_param.name == "ctx":  # Consume 'ctx'
        pass
    elif (
        first_param
    ):  # If first_param was not 'self' and not 'ctx', it's a command option
        param_iter = iter(params.values())  # Reset iterator to include the first param

    for param in param_iter:
        if param.name == "self" or param.name == "ctx":  # Should have been skipped
            continue

        if param.kind == param.VAR_POSITIONAL or param.kind == param.VAR_KEYWORD:
            # *args and **kwargs are not directly supported by slash command options structure.
            # Could raise an error or ignore. For now, ignore.

            continue

        option_name = param.name
        option_description = param_descriptions.get(
            option_name, f"Description for {option_name}"
        )
        meta = option_meta.get(option_name) if option_meta else None

        param_type_hint = param.annotation
        if param_type_hint == inspect.Parameter.empty:
            # Default to string if no type hint, or raise error.
            # Forcing type hints is generally better for slash commands.
            # raise TypeError(f"Option '{option_name}' must have a type hint for slash commands.")
            param_type_hint = str  # Defaulting to string, can be made stricter

        option_type: Optional[ApplicationCommandOptionType] = None
        choices: Optional[List[ApplicationCommandOptionChoice]] = None

        origin = get_origin(param_type_hint)
        args = get_args(param_type_hint)

        if origin is Annotated:
            param_type_hint = args[0]
            for extra in args[1:]:
                if isinstance(extra, OptionMetadata):
                    meta = extra
            origin = get_origin(param_type_hint)
            args = get_args(param_type_hint)

        actual_type_for_mapping = param_type_hint
        is_optional = False

        if origin is Union:  # Handles Optional[T] which is Union[T, NoneType]
            # Filter out NoneType to get the actual type for mapping
            union_types = [t for t in args if t is not type(None)]
            if len(union_types) == 1:
                actual_type_for_mapping = union_types[0]
                is_optional = True
            else:
                # More complex Unions are not directly supported by a single option type.
                # Could default to STRING or raise.
                # For now, let's assume simple Optional[T] or direct types.

                actual_type_for_mapping = str

        elif origin is list and len(args) == 1:
            # List[T] is not a direct option type. Discord handles multiple values for some types
            # via repeated options or specific component interactions, not directly in slash command options.
            # This might indicate a need for a different interaction pattern or custom parsing.
            # For now, treat List[str] as a string, others might error or default.

            actual_type_for_mapping = args[
                0
            ]  # Use the inner type for mapping, but this is a simplification.

        if origin is Literal:  # typing.Literal['a', 'b']
            choices = []
            for choice_val in args:
                if not isinstance(choice_val, (str, int, float)):
                    raise TypeError(
                        f"Literal choices for '{option_name}' must be str, int, or float. Got {type(choice_val)}."
                    )
                choices.append(
                    ApplicationCommandOptionChoice(
                        data={"name": str(choice_val), "value": choice_val}
                    )
                )
            # The type of the Literal's arguments determines the option type
            if choices:
                literal_arg_type = type(choices[0].value)
                option_type = _type_mapping.get(literal_arg_type)
                if (
                    not option_type and literal_arg_type is float
                ):  # float maps to NUMBER
                    option_type = ApplicationCommandOptionType.NUMBER

        if not option_type:  # If not determined by Literal
            option_type = _type_mapping.get(actual_type_for_mapping)
            # Special handling for User, Member, Role, Attachment, Channel if not directly in _type_mapping
            # This is a bit crude; a proper registry or isinstance checks would be better.
            if not option_type:
                if (
                    actual_type_for_mapping.__name__ == "User"
                    or actual_type_for_mapping.__name__ == "Member"
                ):
                    option_type = ApplicationCommandOptionType.USER
                elif actual_type_for_mapping.__name__ == "Role":
                    option_type = ApplicationCommandOptionType.ROLE
                elif actual_type_for_mapping.__name__ == "Attachment":
                    option_type = ApplicationCommandOptionType.ATTACHMENT
                elif (
                    inspect.isclass(actual_type_for_mapping)
                    and isinstance(Channel, type)
                    and issubclass(actual_type_for_mapping, cast(type, Channel))
                ):
                    option_type = ApplicationCommandOptionType.CHANNEL

        if not option_type:
            # Fallback or error if type couldn't be mapped

            option_type = ApplicationCommandOptionType.STRING  # Default fallback

        required = (param.default == inspect.Parameter.empty) and not is_optional

        data: Dict[str, Any] = {
            "name": option_name,
            "description": option_description,
            "type": option_type.value,
            "required": required,
            "choices": ([c.to_dict() for c in choices] if choices else None),
        }

        if meta:
            if meta.channel_types is not None:
                data["channel_types"] = meta.channel_types
            if meta.min_value is not None:
                data["min_value"] = meta.min_value
            if meta.max_value is not None:
                data["max_value"] = meta.max_value
            if meta.min_length is not None:
                data["min_length"] = meta.min_length
            if meta.max_length is not None:
                data["max_length"] = meta.max_length
            if meta.autocomplete:
                data["autocomplete"] = True

        options.append(ApplicationCommandOption(data=data))
    return options


def _app_command_decorator(
    cls: Type[AppCmdType],
    option_meta: Optional[Dict[str, OptionMetadata]] = None,
    **attrs: Any,
) -> Callable[[Callable[..., Any]], AppCmdType]:
    """Generic factory for creating app command decorators."""

    def decorator(func: Callable[..., Any]) -> AppCmdType:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(
                "Application command callback must be a coroutine function."
            )

        name = attrs.pop("name", None) or func.__name__
        description = attrs.pop("description", None) or inspect.getdoc(func)
        if description:  # Clean up docstring
            description = inspect.cleandoc(description).split("\n\n", 1)[
                0
            ]  # Use first paragraph

        parent_group = attrs.pop("parent", None)
        if parent_group and not isinstance(parent_group, AppCommandGroup):
            raise TypeError(
                "The 'parent' argument must be an AppCommandGroup instance."
            )

        # For User/Message commands, description should be empty for payload, but can be stored for help.
        if cls is UserCommand or cls is MessageCommand:
            actual_description_for_payload = ""
        else:
            actual_description_for_payload = description
            if not actual_description_for_payload and cls is SlashCommand:
                raise ValueError(f"Slash command '{name}' must have a description.")

        # Create the command instance
        cmd_instance = cls(
            callback=func,
            name=name,
            description=actual_description_for_payload,  # Use payload-appropriate description
            **attrs,  # Remaining attributes like guild_ids, nsfw, etc.
        )

        # Store original description if different (e.g. for User/Message commands for help text)
        if description != actual_description_for_payload:
            cmd_instance._full_description = (
                description  # Custom attribute for library use
            )

        if isinstance(cmd_instance, SlashCommand):
            cmd_instance.options = _extract_options_from_signature(func, option_meta)

        if parent_group:
            parent_group.add_command(cmd_instance)  # This also sets cmd_instance.parent

        # Attach command object to the function for later collection by Cog or Client
        # This is a common pattern.
        if hasattr(func, "__app_command_object__"):
            # Function might already be decorated (e.g. hybrid or stacked decorators)
            # Decide on behavior: error, overwrite, or store list of commands.
            # For now, let's assume one app command decorator of a specific type per function.
            # Hybrid commands will need special handling.
            print(
                f"Warning: Function {func.__name__} is already an app command or has one attached. Overwriting."
            )

        setattr(func, "__app_command_object__", cmd_instance)
        setattr(cmd_instance, "__app_command_object__", cmd_instance)

        # If the command is a HybridCommand, also set the attribute
        # that the prefix command system's Cog._inject looks for.
        if isinstance(cmd_instance, HybridCommand):
            setattr(func, "__command_object__", cmd_instance)
            setattr(cmd_instance, "__command_object__", cmd_instance)

        return cmd_instance  # Return the command instance itself, not the function
        # This allows it to be added to cogs/handlers directly.

    return decorator


def slash_command(
    name: Optional[str] = None,
    description: Optional[str] = None,
    guild_ids: Optional[List[Snowflake]] = None,
    default_member_permissions: Optional[str] = None,
    nsfw: bool = False,
    name_localizations: Optional[Dict[str, str]] = None,
    description_localizations: Optional[Dict[str, str]] = None,
    integration_types: Optional[List[IntegrationType]] = None,
    contexts: Optional[List[InteractionContextType]] = None,
    *,
    guilds: bool = True,
    dms: bool = True,
    private_channels: bool = True,
    parent: Optional[AppCommandGroup] = None,  # Added parent parameter
    locale: Optional[str] = None,
    option_meta: Optional[Dict[str, OptionMetadata]] = None,
) -> Callable[[Callable[..., Any]], SlashCommand]:
    """
    Decorator to create a CHAT_INPUT (slash) command.
    Options are inferred from the function's type hints.
    """
    if contexts is None:
        ctxs: List[InteractionContextType] = []
        if guilds:
            ctxs.append(InteractionContextType.GUILD)
        if dms:
            ctxs.append(InteractionContextType.BOT_DM)
        if private_channels:
            ctxs.append(InteractionContextType.PRIVATE_CHANNEL)
        if len(ctxs) != 3:
            contexts = ctxs
    attrs = {
        "name": name,
        "description": description,
        "guild_ids": guild_ids,
        "default_member_permissions": default_member_permissions,
        "nsfw": nsfw,
        "name_localizations": name_localizations,
        "description_localizations": description_localizations,
        "integration_types": integration_types,
        "contexts": contexts,
        "parent": parent,  # Pass parent to attrs
        "locale": locale,
    }
    # Filter out None values to avoid passing them as explicit None to command constructor
    # Keep 'parent' even if None, as _app_command_decorator handles None parent.
    # nsfw default is False, so it's fine if not present and defaults.
    attrs = {k: v for k, v in attrs.items() if v is not None or k in ["nsfw", "parent"]}
    return _app_command_decorator(SlashCommand, option_meta, **attrs)


def user_command(
    name: Optional[str] = None,
    guild_ids: Optional[List[Snowflake]] = None,
    default_member_permissions: Optional[str] = None,
    nsfw: bool = False,  # Though less common for user commands
    name_localizations: Optional[Dict[str, str]] = None,
    integration_types: Optional[List[IntegrationType]] = None,
    contexts: Optional[List[InteractionContextType]] = None,
    locale: Optional[str] = None,
    # description is not used by Discord for User commands
) -> Callable[[Callable[..., Any]], UserCommand]:
    """Decorator to create a USER context menu command."""
    attrs = {
        "name": name,
        "guild_ids": guild_ids,
        "default_member_permissions": default_member_permissions,
        "nsfw": nsfw,
        "name_localizations": name_localizations,
        "integration_types": integration_types,
        "contexts": contexts,
        "locale": locale,
    }
    attrs = {k: v for k, v in attrs.items() if v is not None or k in ["nsfw"]}
    return _app_command_decorator(UserCommand, **attrs)


def message_command(
    name: Optional[str] = None,
    guild_ids: Optional[List[Snowflake]] = None,
    default_member_permissions: Optional[str] = None,
    nsfw: bool = False,  # Though less common for message commands
    name_localizations: Optional[Dict[str, str]] = None,
    integration_types: Optional[List[IntegrationType]] = None,
    contexts: Optional[List[InteractionContextType]] = None,
    locale: Optional[str] = None,
    # description is not used by Discord for Message commands
) -> Callable[[Callable[..., Any]], MessageCommand]:
    """Decorator to create a MESSAGE context menu command."""
    attrs = {
        "name": name,
        "guild_ids": guild_ids,
        "default_member_permissions": default_member_permissions,
        "nsfw": nsfw,
        "name_localizations": name_localizations,
        "integration_types": integration_types,
        "contexts": contexts,
        "locale": locale,
    }
    attrs = {k: v for k, v in attrs.items() if v is not None or k in ["nsfw"]}
    return _app_command_decorator(MessageCommand, **attrs)


def hybrid_command(
    name: Optional[str] = None,
    description: Optional[str] = None,
    guild_ids: Optional[List[Snowflake]] = None,
    default_member_permissions: Optional[str] = None,
    nsfw: bool = False,
    name_localizations: Optional[Dict[str, str]] = None,
    description_localizations: Optional[Dict[str, str]] = None,
    integration_types: Optional[List[IntegrationType]] = None,
    contexts: Optional[List[InteractionContextType]] = None,
    *,
    guilds: bool = True,
    dms: bool = True,
    private_channels: bool = True,
    aliases: Optional[List[str]] = None,  # Specific to prefix command aspect
    # Other prefix-specific options can be added here (e.g., help, brief)
    option_meta: Optional[Dict[str, OptionMetadata]] = None,
    locale: Optional[str] = None,
) -> Callable[[Callable[..., Any]], HybridCommand]:
    """
    Decorator to create a command that can be invoked as both a slash command
    and a traditional prefix-based command.
    Options for the slash command part are inferred from the function's type hints.
    """
    if contexts is None:
        ctxs: List[InteractionContextType] = []
        if guilds:
            ctxs.append(InteractionContextType.GUILD)
        if dms:
            ctxs.append(InteractionContextType.BOT_DM)
        if private_channels:
            ctxs.append(InteractionContextType.PRIVATE_CHANNEL)
        if len(ctxs) != 3:
            contexts = ctxs
    attrs = {
        "name": name,
        "description": description,
        "guild_ids": guild_ids,
        "default_member_permissions": default_member_permissions,
        "nsfw": nsfw,
        "name_localizations": name_localizations,
        "description_localizations": description_localizations,
        "integration_types": integration_types,
        "contexts": contexts,
        "aliases": aliases or [],  # Ensure aliases is a list
        "locale": locale,
    }
    # Filter out None values to avoid passing them as explicit None to command constructor
    # Keep 'nsfw' and 'aliases' as they have defaults (False, [])
    attrs = {
        k: v for k, v in attrs.items() if v is not None or k in ["nsfw", "aliases"]
    }
    return _app_command_decorator(HybridCommand, option_meta, **attrs)


def subcommand(
    parent: AppCommandGroup, *d_args: Any, **d_kwargs: Any
) -> Callable[[Callable[..., Any]], SlashCommand]:
    """Create a subcommand under an existing :class:`AppCommandGroup`."""

    d_kwargs.setdefault("parent", parent)
    return slash_command(*d_args, **d_kwargs)


def group(
    name: str,
    description: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[Optional[Callable[..., Any]]], AppCommandGroup]:
    """Decorator to declare a top level :class:`AppCommandGroup`."""

    def decorator(func: Optional[Callable[..., Any]] = None) -> AppCommandGroup:
        grp = AppCommandGroup(
            name=name,
            description=description,
            guild_ids=kwargs.get("guild_ids"),
            parent=kwargs.get("parent"),
            default_member_permissions=kwargs.get("default_member_permissions"),
            nsfw=kwargs.get("nsfw", False),
            name_localizations=kwargs.get("name_localizations"),
            description_localizations=kwargs.get("description_localizations"),
            integration_types=kwargs.get("integration_types"),
            contexts=kwargs.get("contexts"),
        )

        if func is not None:
            setattr(func, "__app_command_object__", grp)
        return grp

    return decorator


def subcommand_group(
    parent: AppCommandGroup,
    name: str,
    description: Optional[str] = None,
    **kwargs: Any,
) -> Callable[[Optional[Callable[..., Any]]], AppCommandGroup]:
    """Create a nested :class:`AppCommandGroup` under ``parent``."""

    return parent.group(
        name=name,
        description=description,
        **kwargs,
    )



from disagreement.components import component_factory
from disagreement.enums import ComponentType, ButtonStyle


def test_component_factory_button():
    data = {
        "type": ComponentType.BUTTON.value,
        "style": ButtonStyle.PRIMARY.value,
        "label": "Click",
        "custom_id": "x",
    }
    comp = component_factory(data)
    assert comp.to_dict()["label"] == "Click"


def test_component_factory_action_row():
    data = {
        "type": ComponentType.ACTION_ROW.value,
        "components": [
            {
                "type": ComponentType.BUTTON.value,
                "style": ButtonStyle.PRIMARY.value,
                "label": "A",
                "custom_id": "b",
            }
        ],
    }
    row = component_factory(data)
    assert len(row.components) == 1



"""Example moderation bot using the Disagreement library."""

import asyncio
import os
import sys
from typing import Set

# Allow running example from repository root
if os.path.join(os.getcwd(), "examples") == os.path.dirname(os.path.abspath(__file__)):
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import disagreement
from disagreement.ext import commands
from disagreement.models import Member, Message

try:
    from dotenv import load_dotenv
except ImportError:  # pragma: no cover - example helper
    load_dotenv = None
    print("python-dotenv is not installed. Environment variables will not be loaded")

if load_dotenv:
    load_dotenv()

BOT_TOKEN = os.environ.get("DISCORD_BOT_TOKEN")
if not BOT_TOKEN:
    print("DISCORD_BOT_TOKEN environment variable not set")
    sys.exit(1)

intents = (
    disagreement.GatewayIntent.GUILDS
    | disagreement.GatewayIntent.GUILD_MESSAGES
    | disagreement.GatewayIntent.MESSAGE_CONTENT
)
client = disagreement.Client(token=BOT_TOKEN, command_prefix="!", intents=intents)

# Simple list of banned words
BANNED_WORDS: Set[str] = {"badword1", "badword2"}


class ModerationCog(commands.Cog):
    def __init__(self, bot: disagreement.Client) -> None:
        super().__init__(bot)

    @commands.command()
    async def kick(
        self, ctx: commands.CommandContext, member: Member, *, reason: str = ""
    ) -> None:
        """Kick a member from the guild."""
        await member.kick(reason=reason or None)
        await ctx.reply(f"Kicked {member.display_name}")

    @commands.command()
    async def ban(
        self, ctx: commands.CommandContext, member: Member, *, reason: str = ""
    ) -> None:
        """Ban a member from the guild."""
        await member.ban(reason=reason or None)
        await ctx.reply(f"Banned {member.display_name}")


@client.event
async def on_ready() -> None:
    if client.user:
        print(f"Logged in as {client.user.username}#{client.user.discriminator}")
    print("Moderation bot ready.")


@client.event
async def on_message(message: Message) -> None:
    await client._process_message_for_commands(message)
    if message.author.bot:
        return
    content_lower = message.content.lower()
    if any(word in content_lower for word in BANNED_WORDS):
        await message.delete()
        await client.send_message(
            message.channel_id,
            f"{message.author.mention}, your message was removed for inappropriate content.",
        )


async def main() -> None:
    client.add_cog(ModerationCog(client))
    await client.run()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from unittest.mock import AsyncMock
import random

import pytest

from disagreement.gateway import GatewayClient, GatewayException
from disagreement.client import Client


class DummyHTTP:
    async def get_gateway_bot(self):
        return {"url": "ws://example"}

    async def _ensure_session(self):
        self._session = AsyncMock()
        self._session.ws_connect = AsyncMock()


class DummyDispatcher:
    async def dispatch(self, *_):
        pass


class DummyClient:
    def __init__(self):
        self.loop = asyncio.get_running_loop()
        self.application_id = None  # Mock application_id for Client.connect


@pytest.mark.asyncio
async def test_client_connect_backoff(monkeypatch):
    http = DummyHTTP()
    # Mock the GatewayClient's connect method to simulate failures and then success
    mock_gateway_connect = AsyncMock(
        side_effect=[GatewayException("boom"), GatewayException("boom"), None]
    )
    # Create a dummy client instance
    client = Client(
        token="test_token",
        intents=0,
        loop=asyncio.get_running_loop(),
        command_prefix="!",
        verbose=False,
        mention_replies=False,
        shard_count=None,
    )
    # Patch the internal _gateway attribute after client initialization
    # This ensures _initialize_gateway is called and _gateway is set
    await client._initialize_gateway()
    monkeypatch.setattr(client._gateway, "connect", mock_gateway_connect)

    # Mock wait_until_ready to prevent it from blocking the test
    monkeypatch.setattr(client, "wait_until_ready", AsyncMock())

    delays = []

    async def fake_sleep(d):
        delays.append(d)

    monkeypatch.setattr(asyncio, "sleep", fake_sleep)

    # Call the client's connect method, which contains the backoff logic
    await client.connect()

    # Assert that GatewayClient.connect was called the correct number of times
    assert mock_gateway_connect.call_count == 3
    # Assert the delays experienced due to exponential backoff
    assert delays == [5, 10]


@pytest.mark.asyncio
async def test_gateway_reconnect_backoff(monkeypatch):
    http = DummyHTTP()
    dispatcher = DummyDispatcher()
    client = DummyClient()
    gw = GatewayClient(
        http_client=http,
        event_dispatcher=dispatcher,
        token="t",
        intents=0,
        client_instance=client,
        max_retries=3,
        max_backoff=10.0,
    )

    connect_mock = AsyncMock(
        side_effect=[GatewayException("boom"), GatewayException("boom"), None]
    )
    monkeypatch.setattr(gw, "connect", connect_mock)

    delays = []

    async def fake_sleep(d):
        delays.append(d)

    monkeypatch.setattr(asyncio, "sleep", fake_sleep)
    monkeypatch.setattr(random, "uniform", lambda a, b: 0)

    await gw._reconnect()

    assert connect_mock.call_count == 3
    assert delays == [1.0, 2.0]



"""
Application Commands Extension for Disagreement.

This package provides the framework for creating and handling
Discord Application Commands (slash commands, user commands, message commands).
"""

from .commands import (
    AppCommand,
    SlashCommand,
    UserCommand,
    MessageCommand,
    AppCommandGroup,
)
from .decorators import (
    slash_command,
    user_command,
    message_command,
    hybrid_command,
    group,
    subcommand,
    subcommand_group,
    OptionMetadata,
)
from .context import AppCommandContext

# from .handler import AppCommandHandler # Will be imported when defined

__all__ = [
    "AppCommand",
    "SlashCommand",
    "UserCommand",
    "MessageCommand",
    "AppCommandGroup",  # To be defined
    "slash_command",
    "user_command",
    "message_command",
    "hybrid_command",
    "group",
    "subcommand",
    "subcommand_group",
    "OptionMetadata",
    "AppCommandContext",  # To be defined
]

from .hybrid import *



import types
from disagreement.models import Message


def make_message(content: str) -> Message:
    data = {
        "id": "1",
        "channel_id": "c",
        "author": {"id": "2", "username": "u", "discriminator": "0001"},
        "content": content,
        "timestamp": "t",
    }
    return Message(data, client_instance=types.SimpleNamespace())


def test_clean_content_removes_mentions():
    msg = make_message("Hello <@123> <#456> <@&789> world")
    assert msg.clean_content == "Hello world"


def test_clean_content_no_mentions():
    msg = make_message("Just text")
    assert msg.clean_content == "Just text"



import pytest

from disagreement.client import Client


def _add_message(client: Client, message_id: str) -> None:
    data = {
        "id": message_id,
        "channel_id": "c",
        "author": {"id": "u", "username": "u", "discriminator": "0001"},
        "content": "hi",
        "timestamp": "t",
    }
    client.parse_message(data)


def test_client_message_cache_size():
    client = Client(token="t", message_cache_maxlen=1)
    _add_message(client, "1")
    assert client._messages.get("1").id == "1"
    _add_message(client, "2")
    assert client._messages.get("1") is None
    assert client._messages.get("2").id == "2"



import asyncio

import pytest
from types import SimpleNamespace

from disagreement.enums import ButtonStyle
from disagreement.models import SelectOption
from disagreement.ui.button import button, Button
from disagreement.ui.select import select, Select
from disagreement.ui.view import View


@pytest.mark.asyncio
async def test_button_decorator_creates_button():
    @button(label="Hi", custom_id="x")
    async def cb(view, inter):
        pass

    assert isinstance(cb, Button)
    assert cb.label == "Hi"
    view = View()
    view.add_item(cb)
    comps = view.to_components_payload()
    assert comps[0]["components"][0]["custom_id"] == "x"


@pytest.mark.asyncio
async def test_button_decorator_requires_coroutine():
    with pytest.raises(TypeError):
        button()(lambda x, y: None)


@pytest.mark.asyncio
async def test_select_decorator_creates_select():
    options = [SelectOption(label="A", value="a")]

    @select(custom_id="sel", options=options)
    async def cb(view, inter):
        pass

    assert isinstance(cb, Select)
    view = View()
    view.add_item(cb)
    payload = view.to_components_payload()[0]["components"][0]
    assert payload["custom_id"] == "sel"


@pytest.mark.asyncio
async def test_view_dispatch_calls_callback(monkeypatch):
    called = {}

    @button(label="B", custom_id="b")
    async def cb(view, inter):
        called["ok"] = True

    view = View()
    view.add_item(cb)

    class DummyInteraction:
        def __init__(self):
            self.data = SimpleNamespace(custom_id="b")

    interaction = DummyInteraction()
    await view._dispatch(interaction)
    assert called.get("ok")



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.client import Client
from disagreement.models import TextChannel
from disagreement.utils import message_pager


@pytest.mark.asyncio
async def test_message_pager_fetches_until_empty():
    calls = [
        [
            {
                "id": "1",
                "channel_id": "c",
                "author": {"id": "2", "username": "u", "discriminator": "0001"},
                "content": "hi",
                "timestamp": "t",
            }
        ],
        [],
    ]
    http = SimpleNamespace(request=AsyncMock(side_effect=calls))
    client = Client.__new__(Client)
    client._http = http
    from disagreement.models import Message

    client.parse_message = lambda d: Message(d, client_instance=client)
    channel = TextChannel({"id": "c", "type": 0}, client)

    messages = []
    async for m in message_pager(channel):
        messages.append(m)

    assert len(messages) == 1
    http.request.assert_awaited()



import pytest

from disagreement.client import Client
from disagreement.models import TextChannel


@pytest.mark.asyncio
async def test_textchannel_history_delegates(monkeypatch):
    called = {}

    async def fake_pager(channel, *, limit=None, before=None, after=None):
        called["args"] = (channel, limit, before, after)
        if False:
            yield None

    monkeypatch.setattr("disagreement.utils.message_pager", fake_pager)
    client = Client.__new__(Client)
    channel = TextChannel({"id": "c", "type": 0}, client)

    hist = channel.history(limit=2, before="b")
    with pytest.raises(StopAsyncIteration):
        await hist.__anext__()

    assert called["args"] == (channel, 2, "b", None)



import logging
from typing import Optional


def setup_logging(level: int, file: Optional[str] = None) -> None:
    """Configure logging for the library.

    Parameters
    ----------
    level:
        Logging level from the :mod:`logging` module.
    file:
        Optional file path to write logs to. If ``None``, logs are sent to
        standard output.
    """
    handlers: list[logging.Handler] = []
    if file is None:
        handlers.append(logging.StreamHandler())
    else:
        handlers.append(logging.FileHandler(file))

    logging.basicConfig(
        level=level,
        handlers=handlers,
        format="%(asctime)s %(levelname)s %(name)s: %(message)s",
    )



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.client import Client
from disagreement.errors import DisagreementException


@pytest.mark.asyncio
async def test_typing_context_manager_calls_http():
    http = SimpleNamespace(trigger_typing=AsyncMock())
    client = Client.__new__(Client)
    client._http = http
    client._closed = False

    async with client.typing("123"):
        pass

    http.trigger_typing.assert_called_once_with("123")


@pytest.mark.asyncio
async def test_typing_closed():
    http = SimpleNamespace(trigger_typing=AsyncMock())
    client = Client.__new__(Client)
    client._http = http
    client._closed = True

    with pytest.raises(DisagreementException):
        async with client.typing("123"):
            pass


@pytest.mark.asyncio
async def test_context_typing():
    http = SimpleNamespace(trigger_typing=AsyncMock())

    class DummyBot:
        def __init__(self):
            self._http = http
            self._closed = False

        def typing(self, channel_id):
            from disagreement.typing import Typing

            return Typing(self, channel_id)

    bot = DummyBot()
    msg = SimpleNamespace(channel_id="c", id="1", guild_id=None, author=None)

    async def dummy(ctx):
        pass

    from disagreement.ext.commands.core import Command, CommandContext

    cmd = Command(dummy)
    ctx = CommandContext(
        message=msg, bot=bot, prefix="!", command=cmd, invoked_with="dummy"
    )

    async with ctx.typing():
        pass

    http.trigger_typing.assert_called_once_with("c")



"""
HTTP client for interacting with the Discord REST API.
"""

import asyncio
import logging
import aiohttp  # pylint: disable=import-error
import json
from urllib.parse import quote
from typing import Optional, Dict, Any, Union, TYPE_CHECKING, List

from .errors import *  # Import all custom exceptions
from . import __version__  # For User-Agent
from .rate_limiter import RateLimiter
from .interactions import InteractionResponsePayload

if TYPE_CHECKING:
    from .client import Client
    from .models import Message, Webhook, File, StageInstance, Invite
    from .interactions import ApplicationCommand, Snowflake

# Discord API constants
API_BASE_URL = "https://discord.com/api/v10"  # Using API v10

logger = logging.getLogger(__name__)

DISCORD_ERROR_CODE_TO_EXCEPTION = {
    0: GeneralError,
    10001: UnknownAccount,
    10002: UnknownApplication,
    10003: UnknownChannel,
    10004: UnknownGuild,
    10005: UnknownIntegration,
    10006: UnknownInvite,
    10007: UnknownMember,
    10008: UnknownMessage,
    10009: UnknownPermissionOverwrite,
    10010: UnknownProvider,
    10011: UnknownRole,
    10012: UnknownToken,
    10013: UnknownUser,
    10014: UnknownEmoji,
    10015: UnknownWebhook,
    10016: UnknownWebhookService,
    10020: UnknownSession,
    10021: UnknownAsset,
    10026: UnknownBan,
    10027: UnknownSKU,
    10028: UnknownStoreListing,
    10029: UnknownEntitlement,
    10030: UnknownBuild,
    10031: UnknownLobby,
    10032: UnknownBranch,
    10033: UnknownStoreDirectoryLayout,
    10036: UnknownRedistributable,
    10038: UnknownGiftCode,
    10049: UnknownStream,
    10050: UnknownPremiumServerSubscribeCooldown,
    10057: UnknownGuildTemplate,
    10059: UnknownDiscoverableServerCategory,
    10060: UnknownSticker,
    10061: UnknownStickerPack,
    10062: UnknownInteraction,
    10063: UnknownApplicationCommand,
    10065: UnknownVoiceState,
    10066: UnknownApplicationCommandPermissions,
    10067: UnknownStageInstance,
    10068: UnknownGuildMemberVerificationForm,
    10069: UnknownGuildWelcomeScreen,
    10070: UnknownGuildScheduledEvent,
    10071: UnknownGuildScheduledEventUser,
    10087: UnknownTag,
    10097: UnknownSound,
    20001: BotsCannotUseThisEndpoint,
    20002: OnlyBotsCanUseThisEndpoint,
    20009: ExplicitContentCannotBeSentToTheDesiredRecipients,
    20012: NotAuthorizedToPerformThisActionOnThisApplication,
    20016: ActionCannotBePerformedDueToSlowmodeRateLimit,
    20018: OnlyTheOwnerOfThisAccountCanPerformThisAction,
    20022: MessageCannotBeEditedDueToAnnouncementRateLimits,
    20024: UnderMinimumAge,
    20028: ChannelHitWriteRateLimit,
    20029: ServerHitWriteRateLimit,
    20031: DisallowedWordsInStageTopicOrNames,
    20035: GuildPremiumSubscriptionLevelTooLow,
    30001: MaximumNumberOfGuildsReached,
    30002: MaximumNumberOfFriendsReached,
    30003: MaximumNumberOfPinsReached,
    30004: MaximumNumberOfRecipientsReached,
    30005: MaximumNumberOfGuildRolesReached,
    30007: MaximumNumberOfWebhooksReached,
    30008: MaximumNumberOfEmojisReached,
    30010: MaximumNumberOfReactionsReached,
    30011: MaximumNumberOfGroupDMsReached,
    30013: MaximumNumberOfGuildChannelsReached,
    30015: MaximumNumberOfAttachmentsInAMessageReached,
    30016: MaximumNumberOfInvitesReached,
    30018: MaximumNumberOfAnimatedEmojisReached,
    30019: MaximumNumberOfServerMembersReached,
    30030: MaximumNumberOfServerCategoriesReached,
    30031: GuildAlreadyHasATemplate,
    30032: MaximumNumberOfApplicationCommandsReached,
    30033: MaximumNumberOfThreadParticipantsReached,
    30034: MaximumNumberOfDailyApplicationCommandCreatesReached,
    30035: MaximumNumberOfBansForNonGuildMembersExceeded,
    30037: MaximumNumberOfBansFetchesReached,
    30038: MaximumNumberOfUncompletedGuildScheduledEventsReached,
    30039: MaximumNumberOfStickersReached,
    30040: MaximumNumberOfPruneRequestsReached,
    30042: MaximumNumberOfGuildWidgetSettingsUpdatesReached,
    30045: MaximumNumberOfSoundboardSoundsReached,
    30046: MaximumNumberOfEditsToMessagesOlderThan1HourReached,
    30047: MaximumNumberOfPinnedThreadsInAForumChannelReached,
    30048: MaximumNumberOfTagsInAForumChannelReached,
    30052: BitrateIsTooHighForChannelOfThisType,
    30056: MaximumNumberOfPremiumEmojisReached,
    30058: MaximumNumberOfWebhooksPerGuildReached,
    30061: MaximumNumberOfChannelPermissionOverwritesReached,
    30062: TheChannelsForThisGuildAreTooLarge,
    40001: Unauthorized,
    40002: YouNeedToVerifyYourAccount,
    40003: YouAreOpeningDirectMessagesTooFast,
    40004: SendMessagesHasBeenTemporarilyDisabled,
    40005: RequestEntityTooLarge,
    40006: ThisFeatureHasBeenTemporarilyDisabledServerSide,
    40007: TheUserIsBannedFromThisGuild,
    40012: ConnectionHasBeenRevoked,
    40018: OnlyConsumableSKUsCanBeConsumed,
    40019: YouCanOnlyDeleteSandboxEntitlements,
    40032: TargetUserIsNotConnectedToVoice,
    40033: ThisMessageHasAlreadyBeenCrossposted,
    40041: AnApplicationCommandWithThatNameAlreadyExists,
    40043: ApplicationInteractionFailedToSend,
    40058: CannotSendAMessageInAForumChannel,
    40060: InteractionHasAlreadyBeenAcknowledged,
    40061: TagNamesMustBeUnique,
    40062: ServiceResourceIsBeingRateLimited,
    40066: ThereAreNoTagsAvailableThatCanBeSetByNonModerators,
    40067: ATagIsRequiredToCreateAForumPostInThisChannel,
    40074: AnEntitlementHasAlreadyBeenGrantedForThisResource,
    40094: ThisInteractionHasHitTheMaximumNumberOfFollowUpMessages,
    40333: CloudflareIsBlockingYourRequest,
    50001: MissingAccess,
    50002: InvalidAccountType,
    50003: CannotExecuteActionOnADMChannel,
    50004: GuildWidgetDisabled,
    50005: CannotEditAMessageAuthoredByAnotherUser,
    50006: CannotSendAnEmptyMessage,
    50007: CannotSendMessagesToThisUser,
    50008: CannotSendMessagesInANonTextChannel,
    50009: ChannelVerificationLevelIsTooHighForYouToGainAccess,
    50010: OAuth2ApplicationDoesNotHaveABot,
    50011: OAuth2ApplicationLimitReached,
    50012: InvalidOAuth2State,
    50013: YouLackPermissionsToPerformThatAction,
    50014: InvalidAuthenticationTokenProvided,
    50015: NoteWasTooLong,
    50016: ProvidedTooFewOrTooManyMessagesToDelete,
    50017: InvalidMFALevel,
    50019: AMessageCanOnlyBePinnedToTheChannelItWasSentIn,
    50020: InviteCodeWasEitherInvalidOrTaken,
    50021: CannotExecuteActionOnASystemMessage,
    50024: CannotExecuteActionOnThisChannelType,
    50025: InvalidOAuth2AccessTokenProvided,
    50026: MissingRequiredOAuth2Scope,
    50027: InvalidWebhookTokenProvided,
    50028: InvalidRole,
    50033: InvalidRecipients,
    50034: AMessageProvidedWasTooOldToBulkDelete,
    50035: InvalidFormBody,
    50036: AnInviteWasAcceptedToAGuildTheApplicationBotIsNotIn,
    50039: InvalidActivityAction,
    50041: InvalidAPIVersionProvided,
    50045: FileUploadedExceedsTheMaximumSize,
    50046: InvalidFileUploaded,
    50054: CannotSelfRedeemThisGift,
    50055: InvalidGuild,
    50057: InvalidSKU,
    50067: InvalidRequestOrigin,
    50068: InvalidMessageType,
    50070: PaymentSourceRequiredToRedeemGift,
    50073: CannotModifyASystemWebhook,
    50074: CannotDeleteAChannelRequiredForCommunityGuilds,
    50080: CannotEditStickersWithinAMessage,
    50081: InvalidStickerSent,
    50083: TriedToPerformAnOperationOnAnArchivedThread,
    50085: InvalidThreadNotificationSettings,
    50086: BeforeValueIsEarlierThanTheThreadCreationDate,
    50087: CommunityServerChannelsMustBeTextChannels,
    50091: TheEntityTypeOfTheEventIsDifferentFromTheEntityYouAreTryingToStartTheEventFor,
    50095: ThisServerIsNotAvailableInYourLocation,
    50097: ThisServerNeedsMonetizationEnabledInOrderToPerformThisAction,
    50101: ThisServerNeedsMoreBoostsToPerformThisAction,
    50109: TheRequestBodyContainsInvalidJSON,
    50110: TheProvidedFileIsInvalid,
    50123: TheProvidedFileTypeIsInvalid,
    50124: TheProvidedFileDurationExceedsMaximumOf52Seconds,
    50131: OwnerCannotBePendingMember,
    50132: OwnershipCannotBeTransferredToABotUser,
    50138: FailedToResizeAssetBelowTheMaximumSize,
    50144: CannotMixSubscriptionAndNonSubscriptionRolesForAnEmoji,
    50145: CannotConvertBetweenPremiumEmojiAndNormalEmoji,
    50146: UploadedFileNotFound,
    50151: TheSpecifiedEmojiIsInvalid,
    50159: VoiceMessagesDoNotSupportAdditionalContent,
    50160: VoiceMessagesMustHaveASingleAudioAttachment,
    50161: VoiceMessagesMustHaveSupportingMetadata,
    50162: VoiceMessagesCannotBeEdited,
    50163: CannotDeleteGuildSubscriptionIntegration,
    50173: YouCannotSendVoiceMessagesInThisChannel,
    50178: TheUserAccountMustFirstBeVerified,
    50192: TheProvidedFileDoesNotHaveAValidDuration,
    50600: YouDoNotHavePermissionToSendThisSticker,
    60003: TwoFactorIsRequiredForThisOperation,
    80004: NoUsersWithDiscordTagExist,
    90001: ReactionWasBlocked,
    90002: UserCannotUseBurstReactions,
    110001: ApplicationNotYetAvailable,
    130000: APIResourceIsCurrentlyOverloaded,
    150006: TheStageIsAlreadyOpen,
    160002: CannotReplyWithoutPermissionToReadMessageHistory,
    160004: AThreadHasAlreadyBeenCreatedForThisMessage,
    160005: ThreadIsLocked,
    160006: MaximumNumberOfActiveThreadsReached,
    160007: MaximumNumberOfActiveAnnouncementThreadsReached,
    170001: InvalidJSONForUploadedLottieFile,
    170002: UploadedLottiesCannotContainRasterizedImages,
    170003: StickerMaximumFramerateExceeded,
    170004: StickerFrameCountExceedsMaximumOf1000Frames,
    170005: LottieAnimationMaximumDimensionsExceeded,
    170006: StickerFrameRateIsEitherTooSmallOrTooLarge,
    170007: StickerAnimationDurationExceedsMaximumOf5Seconds,
    180000: CannotUpdateAFinishedEvent,
    180002: FailedToCreateStageNeededForStageEvent,
    200000: MessageWasBlockedByAutomaticModeration,
    200001: TitleWasBlockedByAutomaticModeration,
    220001: WebhooksPostedToForumChannelsMustHaveAThreadNameOrThreadId,
    220002: WebhooksPostedToForumChannelsCannotHaveBothAThreadNameAndThreadId,
    220003: WebhooksCanOnlyCreateThreadsInForumChannels,
    220004: WebhookServicesCannotBeUsedInForumChannels,
    240000: MessageBlockedByHarmfulLinksFilter,
    350000: CannotEnableOnboardingRequirementsAreNotMet,
    350001: CannotUpdateOnboardingWhileBelowRequirements,
    500000: FailedToBanUsers,
    520000: PollVotingBlocked,
    520001: PollExpired,
    520002: InvalidChannelTypeForPollCreation,
    520003: CannotEditAPollMessage,
    520004: CannotUseAnEmojiIncludedWithThePoll,
    520006: CannotExpireANonPollMessage,
}


class HTTPClient:
    """Handles HTTP requests to the Discord API."""

    def __init__(
        self,
        token: str,
        client_session: Optional[aiohttp.ClientSession] = None,
        verbose: bool = False,
        **session_kwargs: Any,
    ):
        """Create a new HTTP client.

        Parameters
        ----------
        token:
            Bot token for authentication.
        client_session:
            Optional existing :class:`aiohttp.ClientSession`.
        verbose:
            If ``True``, log HTTP requests and responses.
        **session_kwargs:
            Additional options forwarded to :class:`aiohttp.ClientSession`, such
            as ``proxy`` or ``connector``.
        """

        self.token = token
        self._session: Optional[aiohttp.ClientSession] = client_session
        self._session_kwargs: Dict[str, Any] = session_kwargs
        self.user_agent = f"DiscordBot (https://github.com/Slipstreamm/disagreement, {__version__})"  # Customize URL

        self.verbose = verbose

        self._rate_limiter = RateLimiter()

    async def _ensure_session(self):
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession(**self._session_kwargs)

    async def close(self):
        """Closes the underlying aiohttp.ClientSession."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def request(
        self,
        method: str,
        endpoint: str,
        payload: Optional[
            Union[Dict[str, Any], List[Dict[str, Any]], aiohttp.FormData]
        ] = None,
        params: Optional[Dict[str, Any]] = None,
        is_json: bool = True,
        use_auth_header: bool = True,
        custom_headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        """Makes an HTTP request to the Discord API."""
        await self._ensure_session()

        url = f"{API_BASE_URL}{endpoint}"
        final_headers: Dict[str, str] = {  # Renamed to final_headers
            "User-Agent": self.user_agent,
        }
        if use_auth_header:
            final_headers["Authorization"] = f"Bot {self.token}"

        if is_json and payload:
            final_headers["Content-Type"] = "application/json"

        if custom_headers:  # Merge custom headers
            final_headers.update(custom_headers)

        if self.verbose:
            logger.debug(
                "HTTP REQUEST: %s %s | payload=%s params=%s",
                method,
                url,
                payload,
                params,
            )

        route = f"{method.upper()}:{endpoint}"

        for attempt in range(5):  # Max 5 retries for rate limits
            await self._rate_limiter.acquire(route)
            assert self._session is not None, "ClientSession not initialized"
            async with self._session.request(
                method,
                url,
                json=payload if is_json else None,
                data=payload if not is_json else None,
                headers=final_headers,
                params=params,
            ) as response:

                data = None
                try:
                    if response.headers.get("Content-Type", "").startswith(
                        "application/json"
                    ):
                        data = await response.json()
                    else:
                        # For non-JSON responses, like fetching images or other files
                        # We might return the raw response or handle it differently
                        # For now, let's assume most API calls expect JSON
                        data = await response.text()
                except (aiohttp.ContentTypeError, json.JSONDecodeError):
                    data = (
                        await response.text()
                    )  # Fallback to text if JSON parsing fails

                if self.verbose:
                    logger.debug(
                        "HTTP RESPONSE: %s %s | %s", response.status, url, data
                    )

                self._rate_limiter.release(route, response.headers)

                if 200 <= response.status < 300:
                    if response.status == 204:
                        return None
                    return data

                # Rate limit handling
                if response.status == 429:  # Rate limited
                    retry_after_str = response.headers.get("Retry-After", "1")
                    try:
                        retry_after = float(retry_after_str)
                    except ValueError:
                        retry_after = 1.0  # Default retry if header is malformed

                    is_global = (
                        response.headers.get("X-RateLimit-Global", "false").lower()
                        == "true"
                    )

                    error_message = f"Rate limited on {method} {endpoint}."
                    if data and isinstance(data, dict) and "message" in data:
                        error_message += f" Discord says: {data['message']}"

                    await self._rate_limiter.handle_rate_limit(
                        route, retry_after, is_global
                    )

                    if attempt < 4:  # Don't log on the last attempt before raising
                        logger.warning(
                            "%s Retrying after %ss (Attempt %s/5). Global: %s",
                            error_message,
                            retry_after,
                            attempt + 1,
                            is_global,
                        )
                        continue  # Retry the request
                    else:  # Last attempt failed
                        raise RateLimitError(
                            response,
                            message=error_message,
                            retry_after=retry_after,
                            is_global=is_global,
                        )

                # Other error handling
                if response.status == 401:  # Unauthorized
                    raise AuthenticationError(response, "Invalid token provided.")
                if response.status == 403:  # Forbidden
                    raise HTTPException(
                        response,
                        "Missing permissions or access denied.",
                        status=response.status,
                        text=str(data),
                    )

                # General HTTP error
                error_text = str(data) if data else "Unknown error"
                discord_error_code = (
                    data.get("code") if isinstance(data, dict) else None
                )

                if discord_error_code in DISCORD_ERROR_CODE_TO_EXCEPTION:
                    exc_class = DISCORD_ERROR_CODE_TO_EXCEPTION[discord_error_code]
                    raise exc_class(
                        response,
                        f"API Error on {method} {endpoint}: {error_text}",
                        status=response.status,
                        text=error_text,
                        error_code=discord_error_code,
                    )

                raise HTTPException(
                    response,
                    f"API Error on {method} {endpoint}: {error_text}",
                    status=response.status,
                    text=error_text,
                    error_code=discord_error_code,
                )
        raise DisagreementException(
            f"Failed request to {method} {endpoint} after multiple retries."
        )

    # --- Specific API call methods ---

    async def get_gateway_bot(self) -> Dict[str, Any]:
        """Gets the WSS URL and sharding information for the Gateway."""
        return await self.request("GET", "/gateway/bot")

    async def send_message(
        self,
        channel_id: str,
        content: Optional[str] = None,
        tts: bool = False,
        embeds: Optional[List[Dict[str, Any]]] = None,
        components: Optional[List[Dict[str, Any]]] = None,
        allowed_mentions: Optional[dict] = None,
        message_reference: Optional[Dict[str, Any]] = None,
        attachments: Optional[List[Any]] = None,
        files: Optional[List[Any]] = None,
        flags: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Sends a message to a channel.

        Parameters
        ----------
        attachments:
            A list of attachment payloads to include with the message.
        files:
            A list of :class:`File` objects containing binary data to upload.

        Returns
        -------
        Dict[str, Any]
            The created message data.
        """
        payload: Dict[str, Any] = {}
        if content is not None:  # Content is optional if embeds/components are present
            payload["content"] = content
        if tts:
            payload["tts"] = True
        if embeds:
            payload["embeds"] = embeds
        if components:
            payload["components"] = components
        if allowed_mentions:
            payload["allowed_mentions"] = allowed_mentions
        all_files: List["File"] = []
        if attachments is not None:
            payload["attachments"] = []
            for a in attachments:
                if hasattr(a, "data") and hasattr(a, "filename"):
                    idx = len(all_files)
                    all_files.append(a)
                    payload["attachments"].append({"id": idx, "filename": a.filename})
                else:
                    payload["attachments"].append(
                        a.to_dict() if hasattr(a, "to_dict") else a
                    )
        if files is not None:
            for f in files:
                if hasattr(f, "data") and hasattr(f, "filename"):
                    idx = len(all_files)
                    all_files.append(f)
                    if "attachments" not in payload:
                        payload["attachments"] = []
                    payload["attachments"].append({"id": idx, "filename": f.filename})
                else:
                    raise TypeError("files must be File objects")
        if flags:
            payload["flags"] = flags
        if message_reference:
            payload["message_reference"] = message_reference

        if not payload:
            raise ValueError("Message must have content, embeds, or components.")

        if all_files:
            form = aiohttp.FormData()
            form.add_field(
                "payload_json", json.dumps(payload), content_type="application/json"
            )
            for idx, f in enumerate(all_files):
                form.add_field(
                    f"files[{idx}]",
                    f.data,
                    filename=f.filename,
                    content_type="application/octet-stream",
                )
            return await self.request(
                "POST",
                f"/channels/{channel_id}/messages",
                payload=form,
                is_json=False,
            )

        return await self.request(
            "POST", f"/channels/{channel_id}/messages", payload=payload
        )

    async def edit_message(
        self,
        channel_id: str,
        message_id: str,
        payload: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Edits a message in a channel."""

        return await self.request(
            "PATCH",
            f"/channels/{channel_id}/messages/{message_id}",
            payload=payload,
        )

    async def get_message(
        self, channel_id: "Snowflake", message_id: "Snowflake"
    ) -> Dict[str, Any]:
        """Fetches a message from a channel."""

        return await self.request(
            "GET", f"/channels/{channel_id}/messages/{message_id}"
        )

    async def delete_message(
        self, channel_id: "Snowflake", message_id: "Snowflake"
    ) -> None:
        """Deletes a message in a channel."""

        await self.request("DELETE", f"/channels/{channel_id}/messages/{message_id}")

    async def create_reaction(
        self, channel_id: "Snowflake", message_id: "Snowflake", emoji: str
    ) -> None:
        """Adds a reaction to a message as the current user."""
        encoded = quote(emoji)
        await self.request(
            "PUT",
            f"/channels/{channel_id}/messages/{message_id}/reactions/{encoded}/@me",
        )

    async def delete_reaction(
        self, channel_id: "Snowflake", message_id: "Snowflake", emoji: str
    ) -> None:
        """Removes the current user's reaction from a message."""
        encoded = quote(emoji)
        await self.request(
            "DELETE",
            f"/channels/{channel_id}/messages/{message_id}/reactions/{encoded}/@me",
        )

    async def delete_user_reaction(
        self,
        channel_id: "Snowflake",
        message_id: "Snowflake",
        emoji: str,
        user_id: "Snowflake",
    ) -> None:
        """Removes another user's reaction from a message."""
        encoded = quote(emoji)
        await self.request(
            "DELETE",
            f"/channels/{channel_id}/messages/{message_id}/reactions/{encoded}/{user_id}",
        )

    async def get_reactions(
        self, channel_id: "Snowflake", message_id: "Snowflake", emoji: str
    ) -> List[Dict[str, Any]]:
        """Fetches the users that reacted with a specific emoji."""
        encoded = quote(emoji)
        return await self.request(
            "GET",
            f"/channels/{channel_id}/messages/{message_id}/reactions/{encoded}",
        )

    async def clear_reactions(
        self, channel_id: "Snowflake", message_id: "Snowflake"
    ) -> None:
        """Removes all reactions from a message."""

        await self.request(
            "DELETE",
            f"/channels/{channel_id}/messages/{message_id}/reactions",
        )

    async def bulk_delete_messages(
        self, channel_id: "Snowflake", messages: List["Snowflake"]
    ) -> List["Snowflake"]:
        """Bulk deletes messages in a channel and returns their IDs."""

        await self.request(
            "POST",
            f"/channels/{channel_id}/messages/bulk-delete",
            payload={"messages": messages},
        )
        return messages

    async def get_pinned_messages(
        self, channel_id: "Snowflake"
    ) -> List[Dict[str, Any]]:
        """Fetches all pinned messages in a channel."""

        return await self.request("GET", f"/channels/{channel_id}/pins")

    async def pin_message(
        self, channel_id: "Snowflake", message_id: "Snowflake"
    ) -> None:
        """Pins a message in a channel."""

        await self.request("PUT", f"/channels/{channel_id}/pins/{message_id}")

    async def unpin_message(
        self, channel_id: "Snowflake", message_id: "Snowflake"
    ) -> None:
        """Unpins a message from a channel."""

        await self.request("DELETE", f"/channels/{channel_id}/pins/{message_id}")

    async def delete_channel(
        self, channel_id: str, reason: Optional[str] = None
    ) -> None:
        """Deletes a channel.

        If the channel is a guild channel, requires the MANAGE_CHANNELS permission.
        If the channel is a thread, requires the MANAGE_THREADS permission (if locked) or
        be the thread creator (if not locked).
        Deleting a category does not delete its child channels.
        """
        custom_headers = {}
        if reason:
            custom_headers["X-Audit-Log-Reason"] = reason

        await self.request(
            "DELETE",
            f"/channels/{channel_id}",
            custom_headers=custom_headers if custom_headers else None,
        )

    async def edit_channel(
        self,
        channel_id: "Snowflake",
        payload: Dict[str, Any],
        reason: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Edits a channel."""
        headers = {"X-Audit-Log-Reason": reason} if reason else None
        return await self.request(
            "PATCH",
            f"/channels/{channel_id}",
            payload=payload,
            custom_headers=headers,
        )

    async def get_channel(self, channel_id: str) -> Dict[str, Any]:
        """Fetches a channel by ID."""
        return await self.request("GET", f"/channels/{channel_id}")

    async def get_channel_invites(
        self, channel_id: "Snowflake"
    ) -> List[Dict[str, Any]]:
        """Fetches the invites for a channel."""

        return await self.request("GET", f"/channels/{channel_id}/invites")

    async def create_invite(
        self, channel_id: "Snowflake", payload: Dict[str, Any]
    ) -> "Invite":
        """Creates an invite for a channel."""

        data = await self.request(
            "POST", f"/channels/{channel_id}/invites", payload=payload
        )
        from .models import Invite

        return Invite.from_dict(data)

    async def delete_invite(self, code: str) -> None:
        """Deletes an invite by code."""

        await self.request("DELETE", f"/invites/{code}")

    async def create_webhook(
        self, channel_id: "Snowflake", payload: Dict[str, Any]
    ) -> "Webhook":
        """Creates a webhook in the specified channel."""

        data = await self.request(
            "POST", f"/channels/{channel_id}/webhooks", payload=payload
        )
        from .models import Webhook

        return Webhook(data)

    async def edit_webhook(
        self, webhook_id: "Snowflake", payload: Dict[str, Any]
    ) -> "Webhook":
        """Edits an existing webhook."""

        data = await self.request("PATCH", f"/webhooks/{webhook_id}", payload=payload)
        from .models import Webhook

        return Webhook(data)

    async def delete_webhook(self, webhook_id: "Snowflake") -> None:
        """Deletes a webhook."""

        await self.request("DELETE", f"/webhooks/{webhook_id}")

    async def execute_webhook(
        self,
        webhook_id: "Snowflake",
        token: str,
        *,
        content: Optional[str] = None,
        tts: bool = False,
        embeds: Optional[List[Dict[str, Any]]] = None,
        components: Optional[List[Dict[str, Any]]] = None,
        allowed_mentions: Optional[dict] = None,
        attachments: Optional[List[Any]] = None,
        files: Optional[List[Any]] = None,
        flags: Optional[int] = None,
        username: Optional[str] = None,
        avatar_url: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Executes a webhook and returns the created message."""

        payload: Dict[str, Any] = {}
        if content is not None:
            payload["content"] = content
        if tts:
            payload["tts"] = True
        if embeds:
            payload["embeds"] = embeds
        if components:
            payload["components"] = components
        if allowed_mentions:
            payload["allowed_mentions"] = allowed_mentions
        if username:
            payload["username"] = username
        if avatar_url:
            payload["avatar_url"] = avatar_url

        all_files: List["File"] = []
        if attachments is not None:
            payload["attachments"] = []
            for a in attachments:
                if hasattr(a, "data") and hasattr(a, "filename"):
                    idx = len(all_files)
                    all_files.append(a)
                    payload["attachments"].append({"id": idx, "filename": a.filename})
                else:
                    payload["attachments"].append(
                        a.to_dict() if hasattr(a, "to_dict") else a
                    )
        if files is not None:
            for f in files:
                if hasattr(f, "data") and hasattr(f, "filename"):
                    idx = len(all_files)
                    all_files.append(f)
                    if "attachments" not in payload:
                        payload["attachments"] = []
                    payload["attachments"].append({"id": idx, "filename": f.filename})
                else:
                    raise TypeError("files must be File objects")
        if flags:
            payload["flags"] = flags

        if all_files:
            form = aiohttp.FormData()
            form.add_field(
                "payload_json", json.dumps(payload), content_type="application/json"
            )
            for idx, f in enumerate(all_files):
                form.add_field(
                    f"files[{idx}]",
                    f.data,
                    filename=f.filename,
                    content_type="application/octet-stream",
                )
            return await self.request(
                "POST",
                f"/webhooks/{webhook_id}/{token}",
                payload=form,
                is_json=False,
                use_auth_header=False,
            )

        return await self.request(
            "POST",
            f"/webhooks/{webhook_id}/{token}",
            payload=payload,
            use_auth_header=False,
        )

    async def get_user(self, user_id: "Snowflake") -> Dict[str, Any]:
        """Fetches a user object for a given user ID."""
        return await self.request("GET", f"/users/{user_id}")

    async def get_current_user_guilds(self) -> List[Dict[str, Any]]:
        """Returns the guilds the current user is in."""
        return await self.request("GET", "/users/@me/guilds")

    async def get_guild_member(
        self, guild_id: "Snowflake", user_id: "Snowflake"
    ) -> Dict[str, Any]:
        """Returns a guild member object for the specified user."""
        return await self.request("GET", f"/guilds/{guild_id}/members/{user_id}")

    async def kick_member(
        self, guild_id: "Snowflake", user_id: "Snowflake", reason: Optional[str] = None
    ) -> None:
        """Kicks a member from the guild."""
        headers = {"X-Audit-Log-Reason": reason} if reason else None
        await self.request(
            "DELETE",
            f"/guilds/{guild_id}/members/{user_id}",
            custom_headers=headers,
        )

    async def ban_member(
        self,
        guild_id: "Snowflake",
        user_id: "Snowflake",
        *,
        delete_message_seconds: int = 0,
        reason: Optional[str] = None,
    ) -> None:
        """Bans a member from the guild."""
        payload = {}
        if delete_message_seconds:
            payload["delete_message_seconds"] = delete_message_seconds
        headers = {"X-Audit-Log-Reason": reason} if reason else None
        await self.request(
            "PUT",
            f"/guilds/{guild_id}/bans/{user_id}",
            payload=payload if payload else None,
            custom_headers=headers,
        )

    async def timeout_member(
        self,
        guild_id: "Snowflake",
        user_id: "Snowflake",
        *,
        until: Optional[str],
        reason: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Times out a member until the given ISO8601 timestamp."""
        payload = {"communication_disabled_until": until}
        headers = {"X-Audit-Log-Reason": reason} if reason else None
        return await self.request(
            "PATCH",
            f"/guilds/{guild_id}/members/{user_id}",
            payload=payload,
            custom_headers=headers,
        )

    async def get_guild_roles(self, guild_id: "Snowflake") -> List[Dict[str, Any]]:
        """Returns a list of role objects for the guild."""
        return await self.request("GET", f"/guilds/{guild_id}/roles")

    async def get_guild(self, guild_id: "Snowflake") -> Dict[str, Any]:
        """Fetches a guild object for a given guild ID."""
        return await self.request("GET", f"/guilds/{guild_id}")

    async def get_guild_widget(self, guild_id: "Snowflake") -> Dict[str, Any]:
        """Fetches the guild widget settings."""

        return await self.request("GET", f"/guilds/{guild_id}/widget")

    async def edit_guild_widget(
        self, guild_id: "Snowflake", payload: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Edits the guild widget settings."""

        return await self.request(
            "PATCH", f"/guilds/{guild_id}/widget", payload=payload
        )

    async def get_guild_templates(self, guild_id: "Snowflake") -> List[Dict[str, Any]]:
        """Fetches all templates for the given guild."""
        return await self.request("GET", f"/guilds/{guild_id}/templates")

    async def create_guild_template(
        self, guild_id: "Snowflake", payload: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Creates a guild template."""
        return await self.request(
            "POST", f"/guilds/{guild_id}/templates", payload=payload
        )

    async def sync_guild_template(
        self, guild_id: "Snowflake", template_code: str
    ) -> Dict[str, Any]:
        """Syncs a guild template to the guild's current state."""
        return await self.request(
            "PUT",
            f"/guilds/{guild_id}/templates/{template_code}",
        )

    async def delete_guild_template(
        self, guild_id: "Snowflake", template_code: str
    ) -> None:
        """Deletes a guild template."""
        await self.request("DELETE", f"/guilds/{guild_id}/templates/{template_code}")

    async def get_guild_scheduled_events(
        self, guild_id: "Snowflake"
    ) -> List[Dict[str, Any]]:
        """Returns a list of scheduled events for the guild."""

        return await self.request("GET", f"/guilds/{guild_id}/scheduled-events")

    async def get_guild_scheduled_event(
        self, guild_id: "Snowflake", event_id: "Snowflake"
    ) -> Dict[str, Any]:
        """Returns a guild scheduled event."""

        return await self.request(
            "GET", f"/guilds/{guild_id}/scheduled-events/{event_id}"
        )

    async def create_guild_scheduled_event(
        self, guild_id: "Snowflake", payload: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Creates a guild scheduled event."""

        return await self.request(
            "POST", f"/guilds/{guild_id}/scheduled-events", payload=payload
        )

    async def edit_guild_scheduled_event(
        self, guild_id: "Snowflake", event_id: "Snowflake", payload: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Edits a guild scheduled event."""

        return await self.request(
            "PATCH",
            f"/guilds/{guild_id}/scheduled-events/{event_id}",
            payload=payload,
        )

    async def delete_guild_scheduled_event(
        self, guild_id: "Snowflake", event_id: "Snowflake"
    ) -> None:
        """Deletes a guild scheduled event."""

        await self.request("DELETE", f"/guilds/{guild_id}/scheduled-events/{event_id}")

    async def get_audit_logs(
        self, guild_id: "Snowflake", **filters: Any
    ) -> Dict[str, Any]:
        """Fetches audit log entries for a guild."""
        params = {k: v for k, v in filters.items() if v is not None}
        return await self.request(
            "GET",
            f"/guilds/{guild_id}/audit-logs",
            params=params if params else None,
        )

    # Add other methods like:
    # async def get_guild(self, guild_id: str) -> Dict[str, Any]: ...
    # async def create_reaction(self, channel_id: str, message_id: str, emoji: str) -> None: ...
    # etc.
    # --- Application Command Endpoints ---

    # Global Application Commands
    async def get_global_application_commands(
        self, application_id: "Snowflake", with_localizations: bool = False
    ) -> List["ApplicationCommand"]:
        """Fetches all global commands for your application."""
        params = {"with_localizations": str(with_localizations).lower()}
        data = await self.request(
            "GET", f"/applications/{application_id}/commands", params=params
        )
        from .interactions import ApplicationCommand  # Ensure constructor is available

        return [ApplicationCommand(cmd_data) for cmd_data in data]

    async def create_global_application_command(
        self, application_id: "Snowflake", payload: Dict[str, Any]
    ) -> "ApplicationCommand":
        """Creates a new global command."""
        data = await self.request(
            "POST", f"/applications/{application_id}/commands", payload=payload
        )
        from .interactions import ApplicationCommand

        return ApplicationCommand(data)

    async def get_global_application_command(
        self, application_id: "Snowflake", command_id: "Snowflake"
    ) -> "ApplicationCommand":
        """Fetches a specific global command."""
        data = await self.request(
            "GET", f"/applications/{application_id}/commands/{command_id}"
        )
        from .interactions import ApplicationCommand

        return ApplicationCommand(data)

    async def edit_global_application_command(
        self,
        application_id: "Snowflake",
        command_id: "Snowflake",
        payload: Dict[str, Any],
    ) -> "ApplicationCommand":
        """Edits a specific global command."""
        data = await self.request(
            "PATCH",
            f"/applications/{application_id}/commands/{command_id}",
            payload=payload,
        )
        from .interactions import ApplicationCommand

        return ApplicationCommand(data)

    async def delete_global_application_command(
        self, application_id: "Snowflake", command_id: "Snowflake"
    ) -> None:
        """Deletes a specific global command."""
        await self.request(
            "DELETE", f"/applications/{application_id}/commands/{command_id}"
        )

    async def bulk_overwrite_global_application_commands(
        self, application_id: "Snowflake", payload: List[Dict[str, Any]]
    ) -> List["ApplicationCommand"]:
        """Bulk overwrites all global commands for your application."""
        data = await self.request(
            "PUT", f"/applications/{application_id}/commands", payload=payload
        )
        from .interactions import ApplicationCommand

        return [ApplicationCommand(cmd_data) for cmd_data in data]

    # Guild Application Commands
    async def get_guild_application_commands(
        self,
        application_id: "Snowflake",
        guild_id: "Snowflake",
        with_localizations: bool = False,
    ) -> List["ApplicationCommand"]:
        """Fetches all commands for your application for a specific guild."""
        params = {"with_localizations": str(with_localizations).lower()}
        data = await self.request(
            "GET",
            f"/applications/{application_id}/guilds/{guild_id}/commands",
            params=params,
        )
        from .interactions import ApplicationCommand

        return [ApplicationCommand(cmd_data) for cmd_data in data]

    async def create_guild_application_command(
        self,
        application_id: "Snowflake",
        guild_id: "Snowflake",
        payload: Dict[str, Any],
    ) -> "ApplicationCommand":
        """Creates a new guild command."""
        data = await self.request(
            "POST",
            f"/applications/{application_id}/guilds/{guild_id}/commands",
            payload=payload,
        )
        from .interactions import ApplicationCommand

        return ApplicationCommand(data)

    async def get_guild_application_command(
        self,
        application_id: "Snowflake",
        guild_id: "Snowflake",
        command_id: "Snowflake",
    ) -> "ApplicationCommand":
        """Fetches a specific guild command."""
        data = await self.request(
            "GET",
            f"/applications/{application_id}/guilds/{guild_id}/commands/{command_id}",
        )
        from .interactions import ApplicationCommand

        return ApplicationCommand(data)

    async def edit_guild_application_command(
        self,
        application_id: "Snowflake",
        guild_id: "Snowflake",
        command_id: "Snowflake",
        payload: Dict[str, Any],
    ) -> "ApplicationCommand":
        """Edits a specific guild command."""
        data = await self.request(
            "PATCH",
            f"/applications/{application_id}/guilds/{guild_id}/commands/{command_id}",
            payload=payload,
        )
        from .interactions import ApplicationCommand

        return ApplicationCommand(data)

    async def delete_guild_application_command(
        self,
        application_id: "Snowflake",
        guild_id: "Snowflake",
        command_id: "Snowflake",
    ) -> None:
        """Deletes a specific guild command."""
        await self.request(
            "DELETE",
            f"/applications/{application_id}/guilds/{guild_id}/commands/{command_id}",
        )

    async def bulk_overwrite_guild_application_commands(
        self,
        application_id: "Snowflake",
        guild_id: "Snowflake",
        payload: List[Dict[str, Any]],
    ) -> List["ApplicationCommand"]:
        """Bulk overwrites all commands for your application for a specific guild."""
        data = await self.request(
            "PUT",
            f"/applications/{application_id}/guilds/{guild_id}/commands",
            payload=payload,
        )
        from .interactions import ApplicationCommand

        return [ApplicationCommand(cmd_data) for cmd_data in data]

    # --- Interaction Response Endpoints ---
    # Note: These methods return Dict[str, Any] representing the Message data.
    # The caller (e.g., AppCommandHandler) will be responsible for constructing Message models
    # if needed, as Message model instantiation requires a `client_instance`.

    async def create_interaction_response(
        self,
        interaction_id: "Snowflake",
        interaction_token: str,
        payload: Union["InteractionResponsePayload", Dict[str, Any]],
        *,
        ephemeral: bool = False,
    ) -> None:
        """Creates a response to an Interaction.

        Parameters
        ----------
        ephemeral: bool
            Ignored parameter for test compatibility.
        """
        # Interaction responses do not use the bot token in the Authorization header.
        # They are authenticated by the interaction_token in the URL.
        payload_data: Dict[str, Any]
        if isinstance(payload, InteractionResponsePayload):
            payload_data = payload.to_dict()
        else:
            payload_data = payload

        await self.request(
            "POST",
            f"/interactions/{interaction_id}/{interaction_token}/callback",
            payload=payload_data,
            use_auth_header=False,
        )

    async def get_original_interaction_response(
        self, application_id: "Snowflake", interaction_token: str
    ) -> Dict[str, Any]:
        """Gets the initial Interaction response."""
        # This endpoint uses the bot token for auth.
        return await self.request(
            "GET", f"/webhooks/{application_id}/{interaction_token}/messages/@original"
        )

    async def edit_original_interaction_response(
        self,
        application_id: "Snowflake",
        interaction_token: str,
        payload: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Edits the initial Interaction response."""
        return await self.request(
            "PATCH",
            f"/webhooks/{application_id}/{interaction_token}/messages/@original",
            payload=payload,
            use_auth_header=False,
        )  # Docs imply webhook-style auth

    async def delete_original_interaction_response(
        self, application_id: "Snowflake", interaction_token: str
    ) -> None:
        """Deletes the initial Interaction response."""
        await self.request(
            "DELETE",
            f"/webhooks/{application_id}/{interaction_token}/messages/@original",
            use_auth_header=False,
        )  # Docs imply webhook-style auth

    async def create_followup_message(
        self,
        application_id: "Snowflake",
        interaction_token: str,
        payload: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Creates a followup message for an Interaction."""
        # Followup messages are sent to a webhook endpoint.
        return await self.request(
            "POST",
            f"/webhooks/{application_id}/{interaction_token}",
            payload=payload,
            use_auth_header=False,
        )  # Docs imply webhook-style auth

    async def edit_followup_message(
        self,
        application_id: "Snowflake",
        interaction_token: str,
        message_id: "Snowflake",
        payload: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Edits a followup message for an Interaction."""
        return await self.request(
            "PATCH",
            f"/webhooks/{application_id}/{interaction_token}/messages/{message_id}",
            payload=payload,
            use_auth_header=False,
        )  # Docs imply webhook-style auth

    async def delete_followup_message(
        self,
        application_id: "Snowflake",
        interaction_token: str,
        message_id: "Snowflake",
    ) -> None:
        """Deletes a followup message for an Interaction."""
        await self.request(
            "DELETE",
            f"/webhooks/{application_id}/{interaction_token}/messages/{message_id}",
            use_auth_header=False,
        )

    async def trigger_typing(self, channel_id: str) -> None:
        """Sends a typing indicator to the specified channel."""
        await self.request("POST", f"/channels/{channel_id}/typing")

    async def start_stage_instance(
        self, payload: Dict[str, Any], reason: Optional[str] = None
    ) -> "StageInstance":
        """Starts a stage instance."""

        headers = {"X-Audit-Log-Reason": reason} if reason else None
        data = await self.request(
            "POST", "/stage-instances", payload=payload, custom_headers=headers
        )
        from .models import StageInstance

        return StageInstance(data)

    async def edit_stage_instance(
        self,
        channel_id: "Snowflake",
        payload: Dict[str, Any],
        reason: Optional[str] = None,
    ) -> "StageInstance":
        """Edits an existing stage instance."""

        headers = {"X-Audit-Log-Reason": reason} if reason else None
        data = await self.request(
            "PATCH",
            f"/stage-instances/{channel_id}",
            payload=payload,
            custom_headers=headers,
        )
        from .models import StageInstance

        return StageInstance(data)

    async def end_stage_instance(
        self, channel_id: "Snowflake", reason: Optional[str] = None
    ) -> None:
        """Ends a stage instance."""

        headers = {"X-Audit-Log-Reason": reason} if reason else None
        await self.request(
            "DELETE", f"/stage-instances/{channel_id}", custom_headers=headers
        )

    async def get_voice_regions(self) -> List[Dict[str, Any]]:
        """Returns available voice regions."""
        return await self.request("GET", "/voice/regions")

    async def start_thread_from_message(
        self,
        channel_id: "Snowflake",
        message_id: "Snowflake",
        payload: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Starts a new thread from an existing message."""
        return await self.request(
            "POST",
            f"/channels/{channel_id}/messages/{message_id}/threads",
            payload=payload,
        )

    async def start_thread_without_message(
        self, channel_id: "Snowflake", payload: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Starts a new thread that is not attached to a message."""
        return await self.request(
            "POST", f"/channels/{channel_id}/threads", payload=payload
        )

    async def join_thread(self, channel_id: "Snowflake") -> None:
        """Joins the current user to a thread."""
        await self.request("PUT", f"/channels/{channel_id}/thread-members/@me")

    async def leave_thread(self, channel_id: "Snowflake") -> None:
        """Removes the current user from a thread."""
        await self.request("DELETE", f"/channels/{channel_id}/thread-members/@me")



import asyncio
import datetime

import pytest

from disagreement.ext import tasks


class Dummy:
    def __init__(self) -> None:
        self.count = 0

    @tasks.loop(seconds=0.01)
    async def work(self) -> None:
        self.count += 1


@pytest.mark.asyncio
async def test_loop_on_error_callback_called() -> None:
    called = False

    def handler(exc: Exception) -> None:  # pragma: no cover - simple callback
        nonlocal called
        called = True

    @tasks.loop(seconds=0.01, on_error=handler)
    async def failing() -> None:
        raise RuntimeError("fail")

    failing.start()
    await asyncio.sleep(0.03)
    failing.stop()
    assert called


@pytest.mark.asyncio
async def test_loop_time_of_day() -> None:
    run_count = 0

    target_time = (datetime.datetime.now() + datetime.timedelta(seconds=0.05)).time()

    @tasks.loop(time_of_day=target_time)
    async def daily() -> None:
        nonlocal run_count
        run_count += 1

    daily.start()
    await asyncio.sleep(0.1)
    daily.stop()
    assert run_count >= 1


@pytest.mark.asyncio
async def test_loop_runs_and_stops() -> None:
    dummy = Dummy()
    dummy.work.start()  # pylint: disable=no-member
    await asyncio.sleep(0.05)
    dummy.work.stop()  # pylint: disable=no-member
    assert dummy.count >= 2
    assert not dummy.work.running  # pylint: disable=no-member


@pytest.mark.asyncio
async def test_before_after_loop_callbacks() -> None:
    events: list[str] = []

    @tasks.loop(seconds=0.01)
    async def ticker() -> None:
        events.append("tick")

    @ticker.before_loop
    async def before() -> None:  # pragma: no cover - trivial callback
        events.append("before")

    @ticker.after_loop
    async def after() -> None:  # pragma: no cover - trivial callback
        events.append("after")

    ticker.start()
    await asyncio.sleep(0.03)
    ticker.stop()
    await asyncio.sleep(0.01)
    assert events and events[0] == "before"
    assert "after" in events



import asyncio
import logging
import traceback
from typing import Optional

from .logging_config import setup_logging


def setup_global_error_handler(
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> None:
    """Configure a basic global error handler for the provided loop.

    The handler logs unhandled exceptions so they don't crash the bot.
    """
    if loop is None:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

    if not logging.getLogger().hasHandlers():
        setup_logging(logging.ERROR)

    def handle_exception(loop: asyncio.AbstractEventLoop, context: dict) -> None:
        exception = context.get("exception")
        if exception:
            logging.error("Unhandled exception in event loop: %s", exception)
            traceback.print_exception(
                type(exception), exception, exception.__traceback__
            )
        else:
            message = context.get("message")
            logging.error("Event loop error: %s", message)

    loop.set_exception_handler(handle_exception)



import asyncio
import pytest
from hypothesis import given, strategies as st

from disagreement.ext.app_commands.converters import run_converters
from disagreement.enums import ApplicationCommandOptionType
from disagreement.errors import AppCommandOptionConversionError
from conftest import DummyInteraction, DummyClient


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "py_type, option_type, input_value, expected",
    [
        (str, ApplicationCommandOptionType.STRING, "hello", "hello"),
        (int, ApplicationCommandOptionType.INTEGER, "42", 42),
        (bool, ApplicationCommandOptionType.BOOLEAN, "true", True),
        (float, ApplicationCommandOptionType.NUMBER, "3.14", pytest.approx(3.14)),
    ],
)
async def test_basic_type_converters(
    basic_interaction, dummy_client, py_type, option_type, input_value, expected
):
    result = await run_converters(
        basic_interaction, py_type, option_type, input_value, dummy_client
    )
    assert result == expected


@pytest.mark.asyncio
async def test_run_converters_error_cases(basic_interaction, dummy_client):
    with pytest.raises(AppCommandOptionConversionError):
        await run_converters(
            basic_interaction,
            bool,
            ApplicationCommandOptionType.BOOLEAN,
            "maybe",
            dummy_client,
        )

    with pytest.raises(AppCommandOptionConversionError):
        await run_converters(
            basic_interaction,
            list,
            ApplicationCommandOptionType.MENTIONABLE,
            "x",
            dummy_client,
        )


@given(st.text())
def test_string_roundtrip(value):
    interaction = DummyInteraction()
    client = DummyClient()
    result = asyncio.run(
        run_converters(
            interaction,
            str,
            ApplicationCommandOptionType.STRING,
            value,
            client,
        )
    )
    assert result == value


@given(st.integers())
def test_integer_roundtrip(value):
    interaction = DummyInteraction()
    client = DummyClient()
    result = asyncio.run(
        run_converters(
            interaction,
            int,
            ApplicationCommandOptionType.INTEGER,
            str(value),
            client,
        )
    )
    assert result == value


@given(st.booleans())
def test_boolean_roundtrip(value):
    interaction = DummyInteraction()
    client = DummyClient()
    raw = "true" if value else "false"
    result = asyncio.run(
        run_converters(
            interaction,
            bool,
            ApplicationCommandOptionType.BOOLEAN,
            raw,
            client,
        )
    )
    assert result is value


@given(st.floats(allow_nan=False, allow_infinity=False))
def test_number_roundtrip(value):
    interaction = DummyInteraction()
    client = DummyClient()
    result = asyncio.run(
        run_converters(
            interaction,
            float,
            ApplicationCommandOptionType.NUMBER,
            str(value),
            client,
        )
    )
    assert result == pytest.approx(value)



import pytest
import aiohttp
from unittest.mock import MagicMock, patch
from disagreement.errors import (
    HTTPException,
    RateLimitError,
    AppCommandOptionConversionError,
    Forbidden,
    NotFound,
    UnknownAccount,
    MaximumNumberOfGuildsReached,
)
from disagreement.http import HTTPClient


# A fixture to provide an HTTPClient with a mocked session
@pytest.fixture
def http_client():
    # Using a real session and patching the request method is more robust
    client = HTTPClient(token="fake_token")
    yield client

    # Cleanup: close the session after the test
    # This requires making the fixture async or running this in an event loop
    async def close_session():
        if client._session:
            await client.close()

    import asyncio

    try:
        loop = asyncio.get_running_loop()
        loop.run_until_complete(close_session())
    except RuntimeError:
        asyncio.run(close_session())


# Mock aiohttp response
class MockAiohttpResponse:
    def __init__(self, status, json_data, headers=None):
        self.status = status
        self._json_data = json_data
        self.headers = headers or {"Content-Type": "application/json"}

    async def json(self):
        return self._json_data

    async def text(self):
        return str(self._json_data)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "error_code, error_message, expected_exception",
    [
        (10001, "Unknown account", UnknownAccount),
        (30001, "Maximum number of guilds reached", MaximumNumberOfGuildsReached),
    ],
)
async def test_error_code_mapping_raises_correct_exception(
    http_client, error_code, error_message, expected_exception
):
    """
    Tests if the HTTP client correctly raises a specific exception
    based on the Discord error code.
    """
    mock_response = MockAiohttpResponse(
        status=400, json_data={"code": error_code, "message": error_message}
    )

    # Patch the session object to control the response
    with patch("aiohttp.ClientSession") as mock_session_class:
        mock_session_instance = mock_session_class.return_value
        mock_session_instance.request.return_value = mock_response

        # Assert that the correct exception is raised
        with pytest.raises(expected_exception) as excinfo:
            await http_client.request("GET", "/test-endpoint")

    # Optionally, check the exception details
    assert excinfo.value.status == 400
    assert excinfo.value.error_code == error_code
    assert error_message in str(excinfo.value)


def test_http_exception_message():
    exc = HTTPException(message="Bad", status=400)
    assert str(exc) == "HTTP 400: Bad"


def test_rate_limit_error_inherits_httpexception():
    exc = RateLimitError(response=None, retry_after=1.0, is_global=True)
    assert isinstance(exc, HTTPException)
    assert "Rate limited" in str(exc)


def test_app_command_option_conversion_error():
    exc = AppCommandOptionConversionError("bad", option_name="opt", original_value="x")
    assert "opt" in str(exc) and "x" in str(exc)


def test_specific_http_exceptions():
    not_found = NotFound(message="missing", status=404)
    forbidden = Forbidden(message="forbidden", status=403)
    assert isinstance(not_found, HTTPException)
    assert isinstance(forbidden, HTTPException)



from disagreement.color import Color
from disagreement.models import Embed, Container, Component


def test_color_parse():
    assert Color.parse(0x123456).value == 0x123456
    assert Color.parse("#123456").value == 0x123456
    c = Color(0xABCDEF)
    assert Color.parse(c) is c
    assert Color.parse(None) is None
    assert Color.parse((255, 0, 0)).value == 0xFF0000


def test_embed_color_parsing():
    e = Embed({"color": "#FF0000"})
    assert e.color.value == 0xFF0000
    e = Embed({"color": Color(0x00FF00)})
    assert e.color.value == 0x00FF00
    e = Embed({"color": 0x0000FF})
    assert e.color.value == 0x0000FF


def test_container_accent_color_parsing():
    container = Container(components=[], accent_color="#010203")
    assert container.accent_color.value == 0x010203
    container = Container(components=[], accent_color=Color(0x111111))
    assert container.accent_color.value == 0x111111
    container = Container(components=[], accent_color=0x222222)
    assert container.accent_color.value == 0x222222



from __future__ import annotations

from typing import Any, Callable, Coroutine, Optional, List, TYPE_CHECKING
import asyncio

from .item import Item
from .view import View
from ..enums import ComponentType, TextInputStyle
from ..models import ActionRow

if TYPE_CHECKING:  # pragma: no cover - for type hints only
    from ..interactions import Interaction


class TextInput(Item):
    """Represents a text input component inside a modal."""

    def __init__(
        self,
        *,
        label: str,
        custom_id: Optional[str] = None,
        style: TextInputStyle = TextInputStyle.SHORT,
        placeholder: Optional[str] = None,
        required: bool = True,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        row: Optional[int] = None,
    ) -> None:
        super().__init__(type=ComponentType.TEXT_INPUT)
        self.label = label
        self.custom_id = custom_id
        self.style = style
        self.placeholder = placeholder
        self.required = required
        self.min_length = min_length
        self.max_length = max_length
        self._row = row

    def to_dict(self) -> dict[str, Any]:
        payload = {
            "type": ComponentType.TEXT_INPUT.value,
            "style": self.style.value,
            "label": self.label,
            "required": self.required,
        }
        if self.custom_id:
            payload["custom_id"] = self.custom_id
        if self.placeholder:
            payload["placeholder"] = self.placeholder
        if self.min_length is not None:
            payload["min_length"] = self.min_length
        if self.max_length is not None:
            payload["max_length"] = self.max_length
        return payload


def text_input(
    *,
    label: str,
    custom_id: Optional[str] = None,
    style: TextInputStyle = TextInputStyle.SHORT,
    placeholder: Optional[str] = None,
    required: bool = True,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    row: Optional[int] = None,
) -> Callable[[Callable[..., Coroutine[Any, Any, Any]]], TextInput]:
    """Decorator to define a text input callback inside a :class:`Modal`."""

    def decorator(func: Callable[..., Coroutine[Any, Any, Any]]) -> TextInput:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("TextInput callback must be a coroutine function.")

        item = TextInput(
            label=label,
            custom_id=custom_id or func.__name__,
            style=style,
            placeholder=placeholder,
            required=required,
            min_length=min_length,
            max_length=max_length,
            row=row,
        )
        item.callback = func
        return item

    return decorator


class Modal:
    """Represents a modal dialog."""

    def __init__(self, *, title: str, custom_id: str) -> None:
        self.title = title
        self.custom_id = custom_id
        self._children: List[TextInput] = []

        for item in self.__class__.__dict__.values():
            if isinstance(item, TextInput):
                self.add_item(item)

    @property
    def children(self) -> List[TextInput]:
        return self._children

    def add_item(self, item: TextInput) -> None:
        if not isinstance(item, TextInput):
            raise TypeError("Only TextInput items can be added to a Modal.")
        if len(self._children) >= 5:
            raise ValueError("A modal can only have up to 5 text inputs.")
        item._view = None  # Not part of a view but reuse item base
        self._children.append(item)

    def to_components(self) -> List[ActionRow]:
        rows: List[ActionRow] = []
        for child in self.children:
            row = ActionRow(components=[child])
            rows.append(row)
        return rows

    def to_dict(self) -> dict[str, Any]:
        return {
            "title": self.title,
            "custom_id": self.custom_id,
            "components": [r.to_dict() for r in self.to_components()],
        }

    async def callback(
        self, interaction: Interaction
    ) -> None:  # pragma: no cover - default
        pass



import asyncio
import time

import pytest

from disagreement.rate_limiter import RateLimiter


@pytest.mark.asyncio
async def test_route_rate_limit_sleep():
    rl = RateLimiter()
    task = asyncio.create_task(rl.handle_rate_limit("GET:/a", 0.05, False))
    await asyncio.sleep(0)  # ensure task starts
    start = time.monotonic()
    await rl.acquire("GET:/a")
    duration = time.monotonic() - start
    await task
    assert duration >= 0.05


@pytest.mark.asyncio
async def test_global_rate_limit_blocks_all_routes():
    rl = RateLimiter()
    task = asyncio.create_task(rl.handle_rate_limit("GET:/a", 0.05, True))
    await asyncio.sleep(0)
    start = time.monotonic()
    await rl.acquire("POST:/b")
    duration = time.monotonic() - start
    await task
    assert duration >= 0.05



"""
Converters for transforming application command option values.
"""

from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Protocol,
    TypeVar,
    Union,
    TYPE_CHECKING,
)
from disagreement.enums import ApplicationCommandOptionType
from disagreement.errors import (
    AppCommandOptionConversionError,
)  # To be created in disagreement/errors.py

if TYPE_CHECKING:
    from disagreement.interactions import Interaction  # For context if needed
    from disagreement.models import (
        User,
        Member,
        Role,
        Channel,
        Attachment,
    )  # Discord models
    from disagreement.client import Client  # For fetching objects

T = TypeVar("T", covariant=True)


class Converter(Protocol[T]):
    """
    A protocol for classes that can convert an interaction option value to a specific type.
    """

    async def convert(self, interaction: "Interaction", value: Any) -> T:
        """
        Converts the given value to the target type.

        Parameters:
            interaction (Interaction): The interaction context.
            value (Any): The raw value from the interaction option.

        Returns:
            T: The converted value.

        Raises:
            AppCommandOptionConversionError: If conversion fails.
        """
        ...


# Basic Type Converters


class StringConverter(Converter[str]):
    async def convert(self, interaction: "Interaction", value: Any) -> str:
        if not isinstance(value, str):
            raise AppCommandOptionConversionError(
                f"Expected a string, but got {type(value).__name__}: {value}"
            )
        return value


class IntegerConverter(Converter[int]):
    async def convert(self, interaction: "Interaction", value: Any) -> int:
        if not isinstance(value, int):
            try:
                return int(value)
            except (ValueError, TypeError):
                raise AppCommandOptionConversionError(
                    f"Expected an integer, but got {type(value).__name__}: {value}"
                )
        return value


class BooleanConverter(Converter[bool]):
    async def convert(self, interaction: "Interaction", value: Any) -> bool:
        if not isinstance(value, bool):
            if isinstance(value, str):
                if value.lower() == "true":
                    return True
                elif value.lower() == "false":
                    return False
            raise AppCommandOptionConversionError(
                f"Expected a boolean, but got {type(value).__name__}: {value}"
            )
        return value


class NumberConverter(Converter[float]):  # Discord 'NUMBER' type is float
    async def convert(self, interaction: "Interaction", value: Any) -> float:
        if not isinstance(value, (int, float)):
            try:
                return float(value)
            except (ValueError, TypeError):
                raise AppCommandOptionConversionError(
                    f"Expected a number (float), but got {type(value).__name__}: {value}"
                )
        return float(value)  # Ensure it's a float even if int is passed


# Discord Model Converters


class UserConverter(Converter["User"]):
    def __init__(self, client: "Client"):
        self._client = client

    async def convert(self, interaction: "Interaction", value: Any) -> "User":
        if isinstance(value, str):  # Assume it's a user ID
            user_id = value
            # Attempt to get from interaction resolved data first
            if (
                interaction.data
                and interaction.data.resolved
                and interaction.data.resolved.users
            ):
                user_object = interaction.data.resolved.users.get(
                    user_id
                )  # This is already a User object
                if user_object:
                    return user_object  # Return the already parsed User object

            # Fallback to fetching if not in resolved or if interaction has no resolved data
            try:
                user = await self._client.fetch_user(
                    user_id
                )  # fetch_user now also parses and caches
                if user:
                    return user
                raise AppCommandOptionConversionError(
                    f"User with ID '{user_id}' not found.",
                    option_name="user",
                    original_value=value,
                )
            except Exception as e:  # Catch potential HTTP errors from fetch_user
                raise AppCommandOptionConversionError(
                    f"Failed to fetch user '{user_id}': {e}",
                    option_name="user",
                    original_value=value,
                )
        elif (
            isinstance(value, dict) and "id" in value
        ):  # If it's raw user data dict (less common path now)
            return self._client.parse_user(value)  # parse_user handles dict -> User
        raise AppCommandOptionConversionError(
            f"Expected a user ID string or user data dict, got {type(value).__name__}",
            option_name="user",
            original_value=value,
        )


class MemberConverter(Converter["Member"]):
    def __init__(self, client: "Client"):
        self._client = client

    async def convert(self, interaction: "Interaction", value: Any) -> "Member":
        if not interaction.guild_id:
            raise AppCommandOptionConversionError(
                "Cannot convert to Member outside of a guild context.",
                option_name="member",
            )

        if isinstance(value, str):  # Assume it's a user ID
            member_id = value
            # Attempt to get from interaction resolved data first
            if (
                interaction.data
                and interaction.data.resolved
                and interaction.data.resolved.members
            ):
                # The Member object from resolved.members should already be correctly initialized
                # by ResolvedData, including its User part.
                member = interaction.data.resolved.members.get(member_id)
                if member:
                    return (
                        member  # Return the already resolved and parsed Member object
                    )

            # Fallback to fetching if not in resolved
            try:
                member = await self._client.fetch_member(
                    interaction.guild_id, member_id
                )
                if member:
                    return member
                raise AppCommandOptionConversionError(
                    f"Member with ID '{member_id}' not found in guild '{interaction.guild_id}'.",
                    option_name="member",
                    original_value=value,
                )
            except Exception as e:
                raise AppCommandOptionConversionError(
                    f"Failed to fetch member '{member_id}': {e}",
                    option_name="member",
                    original_value=value,
                )
        elif isinstance(value, dict) and "id" in value.get(
            "user", {}
        ):  # If it's already a member data dict
            return self._client.parse_member(value, interaction.guild_id)
        raise AppCommandOptionConversionError(
            f"Expected a member ID string or member data dict, got {type(value).__name__}",
            option_name="member",
            original_value=value,
        )


class RoleConverter(Converter["Role"]):
    def __init__(self, client: "Client"):
        self._client = client

    async def convert(self, interaction: "Interaction", value: Any) -> "Role":
        if not interaction.guild_id:
            raise AppCommandOptionConversionError(
                "Cannot convert to Role outside of a guild context.", option_name="role"
            )

        if isinstance(value, str):  # Assume it's a role ID
            role_id = value
            # Attempt to get from interaction resolved data first
            if (
                interaction.data
                and interaction.data.resolved
                and interaction.data.resolved.roles
            ):
                role_object = interaction.data.resolved.roles.get(
                    role_id
                )  # Should be a Role object
                if role_object:
                    return role_object

            # Fallback to fetching from guild if not in resolved
            # This requires Client to have a fetch_role method or similar
            try:
                # Assuming Client.fetch_role(guild_id, role_id) will be implemented
                role = await self._client.fetch_role(interaction.guild_id, role_id)
                if role:
                    return role
                raise AppCommandOptionConversionError(
                    f"Role with ID '{role_id}' not found in guild '{interaction.guild_id}'.",
                    option_name="role",
                    original_value=value,
                )
            except Exception as e:
                raise AppCommandOptionConversionError(
                    f"Failed to fetch role '{role_id}': {e}",
                    option_name="role",
                    original_value=value,
                )
        elif (
            isinstance(value, dict) and "id" in value
        ):  # If it's already role data dict
            if (
                not interaction.guild_id
            ):  # Should have been caught earlier, but as a safeguard
                raise AppCommandOptionConversionError(
                    "Guild context is required to parse role data.",
                    option_name="role",
                    original_value=value,
                )
            return self._client.parse_role(value, interaction.guild_id)
        # This path is reached if value is not a string (role ID) and not a dict (role data)
        # or if it's a string but all fetching/lookup attempts failed.
        # The final raise AppCommandOptionConversionError should be outside the if/elif for string values.
        # If value was a string, an error should have been raised within the 'if isinstance(value, str)' block.
        # If it wasn't a string or dict, this is the correct place to raise.
        # The previous structure was slightly off, as the final raise was inside the string check block.
        # Let's ensure the final raise is at the correct scope.
        # The current structure seems to imply that if it's not a string, it must be a dict or error.
        # If it's a string and all lookups fail, an error is raised within that block.
        # If it's a dict and parsing fails (or guild_id missing), error raised.
        # If it's neither, this final raise is correct.
        # The "Function with declared return type "Role" must return value on all code paths"
        # error suggests a path where no return or raise happens.
        # This happens if `isinstance(value, str)` is true, but then all internal paths
        # (resolved check, fetch try/except) don't lead to a return or raise *before*
        # falling out of the `if isinstance(value, str)` block.
        # The `raise AppCommandOptionConversionError` at the end of the `if isinstance(value, str)` block
        # (line 156 in previous version) handles the case where a role ID is given but not found.
        # The one at the very end (line 164 in previous) handles cases where value is not str/dict.

        # Corrected structure for the final raise:
        # It should be at the same level as the initial `if isinstance(value, str):`
        # to catch cases where `value` is neither a str nor a dict.
        # However, the current logic within the `if isinstance(value, str):` block
        # ensures a raise if the role ID is not found.
        # The `elif isinstance(value, dict)` handles the dict case.
        # The final `raise` (line 164) is for types other than str or dict.
        # The Pylance error "Function with declared return type "Role" must return value on all code paths"
        # implies that if `value` is a string, and `interaction.data.resolved.roles.get(role_id)` is None,
        # AND `self._client.fetch_role` returns None (which it can), then the
        # `raise AppCommandOptionConversionError` on line 156 is correctly hit.
        # The issue might be that Pylance doesn't see `AppCommandOptionConversionError` as definitively terminating.
        # This is unlikely. Let's re-verify the logic flow.

        # The `raise` on line 156 is correct if role_id is not found after fetching.
        # The `raise` on line 164 is for when `value` is not a str and not a dict.
        # This seems logically sound. The Pylance error might be a misinterpretation or a subtle issue.
        # For now, the duplicated `except` is the primary syntax error.
        # The "must return value on all code paths" often occurs if an if/elif chain doesn't
        # exhaust all possibilities or if a path through a try/except doesn't guarantee a return/raise.
        # In this case, if `value` is a string, it either returns a Role or raises an error.
        # If `value` is a dict, it either returns a Role or raises an error.
        # If `value` is neither, it raises an error. All paths seem covered.
        # The syntax error from the duplicated `except` is the most likely culprit for Pylance's confusion.
        raise AppCommandOptionConversionError(
            f"Expected a role ID string or role data dict, got {type(value).__name__}",
            option_name="role",
            original_value=value,
        )


class ChannelConverter(Converter["Channel"]):
    def __init__(self, client: "Client"):
        self._client = client

    async def convert(self, interaction: "Interaction", value: Any) -> "Channel":
        if isinstance(value, str):  # Assume it's a channel ID
            channel_id = value
            # Attempt to get from interaction resolved data first
            if (
                interaction.data
                and interaction.data.resolved
                and interaction.data.resolved.channels
            ):
                # Resolved channels are PartialChannel. Client.fetch_channel will get the full typed one.
                partial_channel = interaction.data.resolved.channels.get(channel_id)
                if partial_channel:
                    # Client.fetch_channel should handle fetching and parsing to the correct Channel subtype
                    full_channel = await self._client.fetch_channel(partial_channel.id)
                    if full_channel:
                        return full_channel
                    # If fetch_channel returns None even with a resolved ID, it's an issue.
                    raise AppCommandOptionConversionError(
                        f"Failed to fetch full channel for resolved ID '{channel_id}'.",
                        option_name="channel",
                        original_value=value,
                    )

            # Fallback to fetching directly if not in resolved or if resolved fetch failed
            try:
                channel = await self._client.fetch_channel(
                    channel_id
                )  # fetch_channel handles parsing
                if channel:
                    return channel
                raise AppCommandOptionConversionError(
                    f"Channel with ID '{channel_id}' not found.",
                    option_name="channel",
                    original_value=value,
                )
            except Exception as e:
                raise AppCommandOptionConversionError(
                    f"Failed to fetch channel '{channel_id}': {e}",
                    option_name="channel",
                    original_value=value,
                )
        # Raw channel data dicts are not typically provided for slash command options.
        raise AppCommandOptionConversionError(
            f"Expected a channel ID string, got {type(value).__name__}",
            option_name="channel",
            original_value=value,
        )


class AttachmentConverter(Converter["Attachment"]):
    def __init__(
        self, client: "Client"
    ):  # Client might be needed for future enhancements or consistency
        self._client = client

    async def convert(self, interaction: "Interaction", value: Any) -> "Attachment":
        if isinstance(value, str):  # Value is the attachment ID
            attachment_id = value
            if (
                interaction.data
                and interaction.data.resolved
                and interaction.data.resolved.attachments
            ):
                attachment_object = interaction.data.resolved.attachments.get(
                    attachment_id
                )  # This is already an Attachment object
                if attachment_object:
                    return (
                        attachment_object  # Return the already parsed Attachment object
                    )
            raise AppCommandOptionConversionError(
                f"Attachment with ID '{attachment_id}' not found in resolved data.",
                option_name="attachment",
                original_value=value,
            )
        raise AppCommandOptionConversionError(
            f"Expected an attachment ID string, got {type(value).__name__}",
            option_name="attachment",
            original_value=value,
        )


# Converters can be registered dynamically using
# :meth:`disagreement.ext.app_commands.handler.AppCommandHandler.register_converter`.

# Mapping from ApplicationCommandOptionType to default converters
# This will be used by the AppCommandHandler to automatically apply converters
# if no explicit converter is specified for a command option's type hint.
DEFAULT_CONVERTERS: Dict[
    ApplicationCommandOptionType, Callable[..., Converter[Any]]
] = {  # Changed Callable signature
    ApplicationCommandOptionType.STRING: StringConverter,
    ApplicationCommandOptionType.INTEGER: IntegerConverter,
    ApplicationCommandOptionType.BOOLEAN: BooleanConverter,
    ApplicationCommandOptionType.NUMBER: NumberConverter,
    ApplicationCommandOptionType.USER: UserConverter,
    ApplicationCommandOptionType.CHANNEL: ChannelConverter,
    ApplicationCommandOptionType.ROLE: RoleConverter,
    # ApplicationCommandOptionType.MENTIONABLE: MentionableConverter, # Special case, can be User or Role
    ApplicationCommandOptionType.ATTACHMENT: AttachmentConverter,  # Added
}


async def run_converters(
    interaction: "Interaction",
    param_type: Any,  # The type hint of the parameter
    option_type: ApplicationCommandOptionType,  # The Discord option type
    value: Any,
    client: "Client",  # Needed for model converters
) -> Any:
    """
    Runs the appropriate converter for a given parameter type and value.
    This function will be more complex, handling custom converters, unions, optionals etc.
    For now, a basic lookup.
    """
    converter_class_factory = DEFAULT_CONVERTERS.get(option_type)
    if converter_class_factory:
        # Check if the factory needs the client instance
        # This is a bit simplistic; a more robust way might involve inspecting __init__ signature
        # or having converters register their needs.
        if option_type in [
            ApplicationCommandOptionType.USER,
            ApplicationCommandOptionType.CHANNEL,  # Anticipating these
            ApplicationCommandOptionType.ROLE,
            ApplicationCommandOptionType.MENTIONABLE,
            ApplicationCommandOptionType.ATTACHMENT,
        ]:
            converter_instance = converter_class_factory(client=client)
        else:
            converter_instance = converter_class_factory()

        return await converter_instance.convert(interaction, value)

    # Fallback for unhandled types or if direct type matching is needed
    if param_type is str and isinstance(value, str):
        return value
    if param_type is int and isinstance(value, int):
        return value
    if param_type is bool and isinstance(value, bool):
        return value
    if param_type is float and isinstance(value, (float, int)):
        return float(value)

    # If no specific converter, and it's not a basic type match, raise error or return raw
    # For now, let's raise if no converter found for a specific option type
    if option_type in DEFAULT_CONVERTERS:
        pass

    # If it's a model type but no converter yet, this will need to be handled
    # e.g. if param_type is User and option_type is ApplicationCommandOptionType.USER

    raise AppCommandOptionConversionError(
        f"No suitable converter found for option type {option_type.name} "
        f"with value '{value}' to target type {param_type.__name__ if hasattr(param_type, '__name__') else param_type}"
    )



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.client import Client
from disagreement.models import TextChannel


@pytest.mark.asyncio
async def test_textchannel_purge_calls_bulk_delete():
    http = SimpleNamespace(
        request=AsyncMock(return_value=[{"id": "1"}, {"id": "2"}]),
        bulk_delete_messages=AsyncMock(),
    )
    client = Client(token="test")
    client._http = http

    channel = TextChannel({"id": "c", "type": 0}, client)

    deleted = await channel.purge(2)

    http.request.assert_called_once_with(
        "GET", "/channels/c/messages", params={"limit": 2}
    )
    http.bulk_delete_messages.assert_awaited_once_with("c", ["1", "2"])
    assert deleted == ["1", "2"]


@pytest.mark.asyncio
async def test_textchannel_purge_before_param():
    http = SimpleNamespace(
        request=AsyncMock(return_value=[]),
        bulk_delete_messages=AsyncMock(),
    )
    client = Client(token="test")
    client._http = http

    channel = TextChannel({"id": "c", "type": 0}, client)

    deleted = await channel.purge(1, before="b")

    http.request.assert_called_once_with(
        "GET", "/channels/c/messages", params={"limit": 1, "before": "b"}
    )
    http.bulk_delete_messages.assert_not_awaited()
    assert deleted == []



import sys
import types

import pytest

from disagreement.ext import loader


def create_dummy_module(name):
    mod = types.ModuleType(name)
    called = {"setup": False, "teardown": False}

    def setup():
        called["setup"] = True

    def teardown():
        called["teardown"] = True

    mod.setup = setup
    mod.teardown = teardown
    sys.modules[name] = mod
    return called


def test_load_and_unload_extension():
    called = create_dummy_module("dummy_ext")

    module = loader.load_extension("dummy_ext")
    assert module is sys.modules["dummy_ext"]
    assert called["setup"] is True

    loader.unload_extension("dummy_ext")
    assert called["teardown"] is True
    assert "dummy_ext" not in loader._loaded_extensions
    assert "dummy_ext" not in sys.modules


def test_load_extension_twice_raises():
    called = create_dummy_module("repeat_ext")
    loader.load_extension("repeat_ext")
    with pytest.raises(ValueError):
        loader.load_extension("repeat_ext")
    loader.unload_extension("repeat_ext")
    assert called["teardown"] is True


def test_reload_extension(monkeypatch):
    called_first = create_dummy_module("reload_ext")
    loader.load_extension("reload_ext")

    called_second = {"setup": False, "teardown": False}
    module_second = types.ModuleType("reload_ext")

    def setup_second():
        called_second["setup"] = True

    def teardown_second():
        called_second["teardown"] = True

    module_second.setup = setup_second
    module_second.teardown = teardown_second

    def import_stub(name):
        assert name == "reload_ext"
        sys.modules[name] = module_second
        return module_second

    monkeypatch.setattr(loader, "import_module", import_stub)

    loader.reload_extension("reload_ext")

    assert called_first["teardown"] is True
    assert called_second["setup"] is True
    assert loader._loaded_extensions["reload_ext"] is module_second

    loader.unload_extension("reload_ext")
    assert called_second["teardown"] is True



import pytest

from disagreement.ui.view import View
from disagreement.ui.button import Button
from disagreement.ui.select import Select
from disagreement.enums import ButtonStyle, ComponentType
from disagreement.models import SelectOption


def test_basic_layout_keeps_one_item_per_row():
    view = View()
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="a"))
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="b"))

    rows = view.to_components()
    assert len(rows) == 2
    assert all(len(r.components) == 1 for r in rows)


def test_advanced_layout_groups_buttons():
    view = View()
    for i in range(6):
        view.add_item(Button(style=ButtonStyle.PRIMARY, label=str(i)))

    rows = view.layout_components_advanced()
    assert len(rows) == 2
    assert len(rows[0].components) == 5
    assert len(rows[1].components) == 1


def test_advanced_layout_select_separate():
    view = View()
    view.add_item(Select(custom_id="s", options=[SelectOption(label="A", value="a")]))
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="b1"))
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="b2"))

    rows = view.layout_components_advanced()
    assert len(rows) == 2
    assert rows[0].components[0].type == ComponentType.STRING_SELECT
    assert all(c.type == ComponentType.BUTTON for c in rows[1].components)
    assert len(rows[1].components) == 2


def test_advanced_layout_respects_row_attribute():
    view = View()
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="x", row=1))
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="y", row=1))
    view.add_item(Button(style=ButtonStyle.PRIMARY, label="z", row=0))

    rows = view.layout_components_advanced()
    assert len(rows) == 2
    assert len(rows[0].components) == 1
    assert len(rows[1].components) == 2
